from gi.repository import Adw, Gtk, GLib


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/gtk/app_row.ui")
class AppRow(Adw.ActionRow):
	__gtype_name__ = "AppRow"
	gtc = Gtk.Template.Child
	image = gtc()
	eol_package_package_status_icon = gtc()
	eol_runtime_status_icon = gtc()
	pinned_status_icon = gtc()
	masked_status_icon = gtc()
	check_button = gtc()

	def idle_stuff(self):
		if self.package.icon_path:
			self.image.add_css_class("icon-dropshadow")
			self.image.set_from_file(self.package.icon_path)

	def gesture_handler(self, *args):
		if self.on_long_press:
			self.on_long_press(self)

	def __init__(self, package, on_long_press=None, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.package = package
		self.on_long_press = on_long_press
		self.rclick_gesture = Gtk.GestureClick(button=3)
		self.long_press_gesture = Gtk.GestureLongPress()

		# Apply
		GLib.idle_add(lambda *_: self.set_title(GLib.markup_escape_text(package.info["name"])))
		GLib.idle_add(lambda *_: self.set_subtitle(GLib.markup_escape_text(package.info["id"])))
		GLib.idle_add(lambda *_: self.idle_stuff())
		self.add_controller(self.rclick_gesture)
		self.add_controller(self.long_press_gesture)
		if package.info["id"] == "io.github.flattool.Warehouse":
			self.check_button.set_active = lambda *_: None
			self.check_button.set_sensitive(False)

		# Connections
		self.rclick_gesture.connect("released", self.gesture_handler)
		self.long_press_gesture.connect("pressed", self.gesture_handler)



from gi.repository import Adw, Gtk, GLib, Gio
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
from src.remotes_page.remote_row import RemoteRow
from src.remotes_page.add_remote_dialog import AddRemoteDialog
from src.gtk.loading_status import LoadingStatus
import subprocess
import gettext

_ = gettext.gettext


class NewRemoteRow(Adw.ActionRow):
	__gtype_name__ = "NewRemoteRow"

	def idle_stuff(self, *args):
		self.set_title(self.info["title"])
		self.set_subtitle(self.info["description"])
		self.add_suffix(Gtk.Image.new_from_icon_name("warehouse:plus-large-symbolic"))

	def __init__(self, info, **kwargs):
		super().__init__(**kwargs)
		self.info = info
		GLib.idle_add(self.idle_stuff)
		self.set_activatable(True)


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/remotes_page/remotes_page.ui")
class RemotesPage(Adw.NavigationPage):
	# Preselected Remotes
	new_remotes = [
		{
			"title": "AppCenter",
			"name": "appcenter",
			"link": "https://flatpak.elementary.io/repo.flatpakrepo",
			"description": _("The open source, pay-what-you-want app store from elementary"),
		},
		{
			"title": "Flathub",
			"name": "flathub",
			"link": "https://dl.flathub.org/repo/flathub.flatpakrepo",
			"description": _("Central repository of Flatpak applications"),
		},
		{
			"title": "Flathub beta",
			"name": "flathub-beta",
			"link": "https://flathub.org/beta-repo/flathub-beta.flatpakrepo",
			"description": _("Beta builds of Flatpak applications"),
		},
		{
			"title": "Fedora",
			"name": "fedora",
			"link": "oci+https://registry.fedoraproject.org",
			"description": _("Flatpaks packaged by Fedora Linux"),
		},
		{
			"title": "GNOME Nightly",
			"name": "gnome-nightly",
			"link": "https://nightly.gnome.org/gnome-nightly.flatpakrepo",
			"description": _("The latest beta GNOME Apps and Runtimes"),
		},
		{
			"title": "WebKit Developer SDK",
			"name": "webkit-sdk",
			"link": "https://software.igalia.com/flatpak-refs/webkit-sdk.flatpakrepo",
			"description": _("Central repository of the WebKit Developer and Runtime SDK"),
		},
	]

	__gtype_name__ = "RemotesPage"
	gtc = Gtk.Template.Child

	search_button = gtc()
	search_bar = gtc()
	search_entry = gtc()
	toast_overlay = gtc()
	stack = gtc()
	current_remotes_group = gtc()
	show_disabled_button = gtc()
	show_disabled_button_content = gtc()
	new_remotes_group = gtc()
	file_remote_row = gtc()
	custom_remote_row = gtc()
	none_visible = gtc()
	status_stack = gtc()
	loading_view = gtc()
	adding_view = gtc()
	main_view = gtc()

	no_results = gtc()
	no_remotes = gtc()
	content_page = gtc()

	# Referred to in the main window
	# It is used to determine if a new page should be made or not
	# This must be set to the created object from within the class's __init__ method
	instance = None
	page_name = "remotes"

	def start_loading(self):
		self.search_button.set_active(False)
		self.status_stack.set_visible_child(self.loading_view)
		self.total_disabled = 0
		for row in self.current_remote_rows:
			self.current_remotes_group.remove(row)

		self.current_remote_rows.clear()

	def end_loading(self):
		show_disabled = self.show_disabled_button.get_active()
		self.show_disabled_button.set_visible(False)
		total_visible = 0
		for installation, remotes in HostInfo.remotes.items():
			for remote in remotes:
				row = RemoteRow(self, installation, remote)
				self.current_remote_rows.append(row)
				self.current_remotes_group.add(row)
				if row.remote.disabled:
					self.total_disabled += 1
					self.show_disabled_button.set_visible(True)
					if show_disabled:
						total_visible += 1
					else:
						row.set_visible(False)
				else:
					total_visible += 1

		self.none_visible.set_visible(total_visible == 0)

		if len(self.current_remote_rows) == 0:
			self.no_remotes.set_visible(True)
			self.none_visible.set_visible(False)
		else:
			self.no_remotes.set_visible(False)

		GLib.idle_add(lambda *_: self.status_stack.set_visible_child(self.main_view))

	def none_visible_handler(self):
		any_visible = False
		for row in self.current_remote_rows:
			if row.get_visible():
				any_visible = True
				break

		self.none_visible.set_visible(not any_visible)

	def filter_remote(self, row):
		self.filter_setting.set_boolean("show-apps", True)
		self.filter_setting.set_boolean("show-runtimes", True)
		self.filter_setting.set_string("remotes-list", f"{row.remote.name}<>{row.installation};")
		self.filter_setting.reset("runtimes-list")
		packages_page = self.main_window.pages[self.main_window.packages_row]
		packages_page.filters_page.generate_filters()
		packages_page.apply_filters()
		GLib.idle_add(lambda *_: self.main_window.activate_row(self.main_window.packages_row))
		GLib.idle_add(lambda *args: packages_page.packages_toast_overlay.add_toast(Adw.Toast(title=_("Showing all packages from {}").format(row.remote.title))))

	def remove_remote(self, row):
		error = [None]

		def thread(*args):
			install = row.installation
			cmd = ["flatpak", "remote-delete", row.remote.name, "--force"]
			if install == "user" or install == "system":
				cmd.append(f"--{install}")
			else:
				cmd.append(f"--installation={install}")

			try:
				subprocess.run(cmd, check=True, capture_output=True, text=True)
			except subprocess.CalledProcessError as cpe:
				error[0] = cpe.stderr
			except Exception as e:
				error[0] = e

		def callback(*args):
			if error[0]:
				self.toast_overlay.add_toast(ErrorToast(_("Could not remove remote"), str(error[0])).toast)
			else:
				filters_page = HostInfo.main_window.pages[HostInfo.main_window.packages_row].filters_page
				filters_page.settings.reset("remotes-list")
				filters_page.all_remotes_switch.set_active(False)
				# filters_page.packages_page.apply_filters()
				self.main_window.refresh_handler()
				self.toast_overlay.add_toast(Adw.Toast(title=_("Removed {}").format(row.remote.title)))

		def on_response(_, response):
			if response != "continue":
				return

			Gio.Task.new(None, None, callback).run_in_thread(thread)

		dialog = Adw.AlertDialog(
			heading=_("Remove {}?").format(row.remote.title), body=_("Any installed apps from {} will stop receiving updates").format(row.remote.name)
		)
		dialog.add_response("cancel", _("Cancel"))
		dialog.add_response("continue", _("Remove"))
		dialog.set_response_appearance("continue", Adw.ResponseAppearance.DESTRUCTIVE)
		dialog.connect("response", on_response)
		dialog.present(self.main_window)

	def on_search(self, entry):
		text = entry.get_text().lower()
		total = 0
		show_disabled = self.show_disabled_button.get_active()

		for row in self.current_remote_rows:
			title_match = text in row.get_title().lower()
			subtitle_match = text in row.get_subtitle().lower()
			visible = (title_match or subtitle_match) and (show_disabled or not row.remote.disabled)
			total += visible
			row.set_visible(visible)

		if text == "":
			self.stack.set_visible_child(self.content_page)
			return

		self.stack.set_visible_child(self.content_page if total > 0 else self.no_results)

	def local_file_handler(self, path):
		try:
			name = path.split("/")[-1].split(".")[0]
			info = {
				"title": name.title(),
				"name": name,
				"description": "local file",
				"link": path,
			}
			AddRemoteDialog(self.main_window, self, info).present(self.main_window)
		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not open file"), str(e)).toast)

	def file_callback(self, chooser, result):
		try:
			file = chooser.open_finish(result)
			path = file.get_path()
			self.local_file_handler(path)
		except GLib.GError as ge:
			if "Dismissed by user" in str(ge):
				return
			self.toast_overlay.add_toast(ErrorToast(_("Could not open file"), str(ge)).toast)
		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not open file"), str(e)).toast)

	def add_file_handler(self):
		file_filter = Gtk.FileFilter(name=_("Flatpak Repos"))
		file_filter.add_suffix("flatpakrepo")
		filters = Gio.ListStore.new(Gtk.FileFilter)
		filters.append(file_filter)
		file_chooser = Gtk.FileDialog()
		file_chooser.set_filters(filters)
		file_chooser.set_default_filter(file_filter)
		file_chooser.open(self.main_window, None, self.file_callback)

	def show_disabled_handler(self, button):
		show_disabled = button.get_active()
		self.show_disabled_button_content.set_icon_name(
			"warehouse:eye-open-negative-filled-symbolic" if show_disabled else "warehouse:eye-not-looking-symbolic"
		)
		total_visible = 0
		for row in self.current_remote_rows:
			if row.remote.disabled:
				if show_disabled:  # show disabled
					row.set_visible(True)
					total_visible += 1
				else:
					row.set_visible(False)
			else:
				total_visible += 1

		self.none_visible.set_visible(total_visible == 0)

	def new_custom_handler(self, *args):
		AddRemoteDialog(self.main_window, self).present(self.main_window)

	def __init__(self, main_window, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.__class__.instance = self
		self.main_window = main_window
		self.search_bar.set_key_capture_widget(main_window)
		self.current_remote_rows = []
		self.filter_setting = Gio.Settings.new("io.github.flattool.Warehouse.filter")
		self.total_disabled = 0

		# Connections
		self.file_remote_row.connect("activated", lambda *_: self.add_file_handler())
		self.custom_remote_row.connect("activated", self.new_custom_handler)
		self.search_entry.connect("search-changed", self.on_search)
		self.show_disabled_button.connect("toggled", self.show_disabled_handler)

		# Apply
		self.adding_view.set_content(LoadingStatus(_("Adding Remote"), _("This should only take a moment")))
		self.loading_view.set_content(LoadingStatus(_("Loading Remotes"), _("This should only take a moment")))
		for item in self.new_remotes:
			row = NewRemoteRow(item)
			row.connect("activated", lambda *_, remote_info=item: AddRemoteDialog(main_window, self, remote_info).present(main_window))
			self.new_remotes_group.add(row)



from gi.repository import Adw, Gtk, GLib, Gio
from src.host_info import HostInfo
from src.gtk.app_row import AppRow
from src.gtk.error_toast import ErrorToast
from src.properties_page.properties_page import PropertiesPage
from src.packages_page.filters_page import FiltersPage
from src.gtk.sidebar_button import SidebarButton
from .uninstall_dialog import UninstallDialog
from src.gtk.loading_status import LoadingStatus
from src.package_install_worker import PackageInstallWorker
from src.change_version_page.change_version_worker import ChangeVersionWorker
import subprocess, os
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/packages_page/packages_page.ui")
class PackagesPage(Adw.BreakpointBin):
	__gtype_name__ = "PackagesPage"
	gtc = Gtk.Template.Child
	packages_bpt = gtc()
	packages_toast_overlay = gtc()
	stack = gtc()
	status_stack = gtc()
	scrolled_window = gtc()
	loading_view = gtc()
	uninstalling_view = gtc()
	reinstalling_view = gtc()
	changing_version_view = gtc()
	no_filter_results = gtc()
	reset_filters_button = gtc()
	no_packages = gtc()
	no_results = gtc()
	filter_button = gtc()
	search_button = gtc()
	search_bar = gtc()
	search_entry = gtc()
	packages_split = gtc()
	packages_list_box = gtc()
	select_button = gtc()
	packages_navpage = gtc()
	select_all_button = gtc()
	content_stack = gtc()
	copy_button = gtc()
	copy_pop = gtc()
	copy_menu = gtc()
	copy_names = gtc()
	copy_ids = gtc()
	copy_refs = gtc()
	uninstall_button = gtc()
	properties_page = gtc()
	filters_page = gtc()

	# Referred to in the main window
	# It is used to determine if a new page should be made or not
	# This must be set to the created object from within the class's __init__ method
	instance = None
	page_name = "packages"
	last_activated_row = None

	def set_status(self, to_set):
		if to_set is self.scrolled_window:
			self.properties_page.stack.set_visible_child(self.properties_page.nav_view)
			self.select_button.set_sensitive(True)
			self.filter_button.set_sensitive(True)
			self.filters_page.set_sensitive(True)

			self.search_button.set_sensitive(True)
			self.search_entry.set_editable(True)
		else:
			self.select_button.set_sensitive(False)

		if to_set is self.no_packages:
			self.properties_page.stack.set_visible_child(self.properties_page.error_tbv)
			self.filter_button.set_sensitive(False)
			self.filter_button.set_active(False)

		if to_set is self.no_filter_results:
			self.properties_page.stack.set_visible_child(self.properties_page.error_tbv)
			self.filter_button.set_sensitive(True)
			self.filters_page.set_sensitive(True)
			if not self.packages_split.get_collapsed():
				self.filter_button.set_active(True)

		if to_set is self.no_results:
			self.filters_page.set_sensitive(False)

		if to_set is self.loading_packages:
			self.stack.set_visible_child(self.loading_view)
		elif to_set is self.uninstalling:
			self.stack.set_visible_child(self.uninstalling_view)
		elif to_set is self.reinstalling:
			self.stack.set_visible_child(self.reinstalling_view)
		elif to_set is self.changing_version:
			self.stack.set_visible_child(self.changing_version_view)
		else:
			self.stack.set_visible_child(self.packages_split)
			self.status_stack.set_visible_child(to_set)

	def apply_filters(self):
		i = 0
		show_apps = self.filter_settings.get_boolean("show-apps")
		show_runtimes = self.filter_settings.get_boolean("show-runtimes")
		remotes_list = self.filter_settings.get_string("remotes-list")
		runtimes_list = self.filter_settings.get_string("runtimes-list")
		total_visible = 0
		while row := self.packages_list_box.get_row_at_index(i):
			i += 1
			visible = True
			if row.package.is_runtime and not show_runtimes:
				visible = False
			if (not row.package.is_runtime) and (not show_apps):
				visible = False
			if remotes_list != "all" and not f"{row.package.info['origin']}<>{row.package.info['installation']}" in remotes_list:
				visible = False
			if runtimes_list != "all" and (
				row.package.is_runtime or row.package.dependent_runtime and not row.package.dependent_runtime.info["ref"] in runtimes_list
			):
				visible = False

			row.set_visible(visible)
			if visible:
				total_visible += 1
			else:
				row.check_button.set_active(False)

		if total_visible == 0:
			self.set_status(self.no_filter_results)
		else:
			GLib.idle_add(lambda *_: self.set_status(self.scrolled_window))
			if self.current_row_for_properties and not self.current_row_for_properties.get_visible():
				self.select_first_visible_row()

	def select_first_visible_row(self):
		first_visible_row = None
		i = 0
		while row := self.packages_list_box.get_row_at_index(i):
			i += 1
			if row.get_visible():
				first_visible_row = row
				self.current_row_for_properties = row
				break

		if not first_visible_row is None:
			self.packages_list_box.select_row(first_visible_row)
			self.properties_page.set_properties(first_visible_row.package)

	def row_select_handler(self, row):
		if row.check_button.get_active():
			self.selected_rows.append(row)
		else:
			self.selected_rows.remove(row)

		if (total := len(self.selected_rows)) > 0:
			self.packages_navpage.set_title(_("{} Selected").format(total))
			self.copy_button.set_sensitive(True)
			self.uninstall_button.set_sensitive(True)
		else:
			self.packages_navpage.set_title(_("Packages"))
			self.copy_button.set_sensitive(False)
			self.uninstall_button.set_sensitive(False)

	def select_all_handler(self, *args):
		i = 0
		while row := self.packages_list_box.get_row_at_index(i):
			i += 1
			row.check_button.set_active(row.get_visible())

	def row_rclick_handler(self, row):
		self.select_button.set_active(True)
		GLib.idle_add(lambda *_, button=row.check_button: button.set_active(not button.get_active()))

	def generate_list(self, *args):
		self.properties_page.nav_view.pop_to_page(self.properties_page.inner_nav_page)
		self.packages_list_box.remove_all()
		self.selected_rows.clear()
		GLib.idle_add(lambda *_: self.filters_page.generate_filters())
		self.copy_button.set_sensitive(False)
		self.uninstall_button.set_sensitive(False)
		if len(HostInfo.flatpaks) == 0:
			self.set_status(self.no_packages)
			return

		for package in HostInfo.flatpaks:
			row = AppRow(package, self.row_rclick_handler)
			package.app_row = row
			row.masked_status_icon.set_visible(package.is_masked)
			row.pinned_status_icon.set_visible(package.is_pinned)
			row.eol_package_package_status_icon.set_visible(package.is_eol)
			row.check_button.set_visible(self.select_button.get_active())
			row.check_button.connect("toggled", lambda *_, row=row: self.row_select_handler(row))
			try:
				if not package.is_runtime:
					row.eol_runtime_status_icon.set_visible(package.dependent_runtime.is_eol)
			except Exception as e:
				self.packages_toast_overlay.add_toast(ErrorToast(_("Error getting Flatpak '{}'").format(package.info["name"]), str(e)).toast)

			self.packages_list_box.append(row)

		self.apply_filters()
		self.select_first_visible_row()

		self.scrolled_window.set_vadjustment(Gtk.Adjustment.new(0, 0, 0, 0, 0, 0))  # Scroll list to top

	def row_activate_handler(self, list_box, row):
		if self.select_button.get_active():
			row.check_button.set_active(not row.check_button.get_active())
			return

		self.last_activated_row = row
		self.properties_page.set_properties(row.package)
		self.properties_page.nav_view.pop()
		self.packages_split.set_show_content(True)
		self.filter_button.set_active(False)
		self.current_row_for_properties = row

	def filter_func(self, row):
		search_text = self.search_entry.get_text().lower()
		title = row.get_title().lower()
		subtitle = row.get_subtitle().lower()
		if row.get_visible() and (search_text in title or search_text in subtitle):
			self.is_result = True
			return True

	def set_selection_mode(self, is_enabled):
		if is_enabled:
			self.packages_list_box.set_selection_mode(Gtk.SelectionMode.NONE)
		else:
			self.packages_list_box.set_selection_mode(Gtk.SelectionMode.SINGLE)
			self.packages_list_box.select_row(self.last_activated_row)

		i = 0
		while row := self.packages_list_box.get_row_at_index(i):
			i += 1
			GLib.idle_add(row.check_button.set_active, False)
			GLib.idle_add(row.check_button.set_visible, is_enabled)

	def selection_copy(self, box, row):
		self.copy_pop.popdown()
		info = ""
		feedback = ""
		match row.get_child():
			case self.copy_names:
				info = "name"
				feedback = _("Names")
			case self.copy_ids:
				info = "id"
				feedback = _("IDs")
			case self.copy_refs:
				info = "ref"
				feedback = _("Refs")

		to_copy = []
		for row in self.selected_rows:
			to_copy.append(row.package.info[info])
			to_copy += ["\n"]
		try:
			HostInfo.clipboard.set("".join(to_copy[:-1]))
			self.packages_toast_overlay.add_toast(Adw.Toast(title=_("Copied {}").format(feedback)))
		except Exception as e:
			self.packages_toast_overlay.add_toast(ErrorToast(_("Could not copy {}").format(feedback), str(e)).toast)

	def selection_uninstall(self, *args):
		if len(self.selected_rows) < 1 or not self.uninstall_button.get_sensitive():
			return

		def on_response(should_trash):
			GLib.idle_add(lambda *_: self.set_status(self.uninstalling))
			error = []

			def thread(*args):
				HostInfo.main_window.add_refresh_lockout("batch uninstalling packages")
				cmd = ["flatpak", "uninstall", "-y"]
				to_uninstall = {}  # { <remote><><installation>: [<ref1>, <ref2>, <ref3>, ...], ... }
				to_trash = []

				for row in self.selected_rows:
					key = row.package.info["installation"]
					if ls := to_uninstall.get(key, False):
						ls.append(row.package.info["ref"])
					else:
						to_uninstall[key] = [row.package.info["ref"]]

					if should_trash and os.path.exists(row.package.data_path):
						to_trash.append(row.package.data_path)

				for installation, packages in to_uninstall.items():
					suffix = []
					if installation == "user" or installation == "system":
						suffix.append(f"--{installation}")
					else:
						suffix.append(f"--installation={installation}")

					try:
						subprocess.run(cmd + suffix + packages, check=True, text=True, capture_output=True)
					except subprocess.CalledProcessError as cpe:
						error.append(str(cpe.stderr))
					except Exception as e:
						error.append(str(e))

				if should_trash and len(to_trash) > 0:
					try:
						subprocess.run(["gio", "trash"] + to_trash, check=True, text=True, capture_output=True)
					except subprocess.CalledProcessError as cpe:
						error.append(cpe)

			def callback(*args):
				self.main_window.refresh_handler()
				HostInfo.main_window.remove_refresh_lockout("batch uninstalling packages")
				if len(error) > 0:
					details = "\n\n".join(error)
					GLib.idle_add(lambda *args: self.packages_toast_overlay.add_toast(ErrorToast(_("Errors occurred while uninstalling"), details).toast))
				else:
					GLib.idle_add(lambda *args: self.packages_toast_overlay.add_toast(Adw.Toast(title=_("Uninstalled Packages"))))

			Gio.Task.new(None, None, callback).run_in_thread(thread)

		dialog = UninstallDialog(on_response, True)
		dialog.present(self.main_window)

	def start_loading(self):
		self.search_button.set_active(False)
		self.last_activated_row = None
		self.packages_navpage.set_title(_("Packages"))
		self.select_button.set_active(False)
		self.set_status(self.loading_packages)

	def end_loading(self):
		GLib.idle_add(lambda *_: self.generate_list())

	def select_button_handler(self, button):
		self.set_selection_mode(button.get_active())

	def filter_button_handler(self, button):
		if button.get_active():
			self.content_stack.set_visible_child(self.filters_page)
			self.packages_split.set_show_content(True)
		else:
			self.content_stack.set_visible_child(self.properties_page)
			self.packages_split.set_show_content(False)

	def filter_page_handler(self, *args):
		if self.packages_split.get_collapsed() and not self.packages_split.get_show_content():
			self.filter_button.set_active(False)

	def on_invalidate(self, row):
		current_status = self.status_stack.get_visible_child()
		if not current_status is self.no_results:
			self.prev_status = current_status

		self.is_result = False
		self.packages_list_box.invalidate_filter()
		if self.is_result:
			self.set_status(self.prev_status)
		else:
			self.set_status(self.no_results)

	def sort_func(self, row1, row2):
		return row1.package.info["name"].lower() > row2.package.info["name"].lower()

	def on_escape_handler(self):
		if self.select_button.get_active():
			self.select_button.set_active(False)
		elif self.filter_button.get_active():
			self.filter_button.set_active(False)

	def __init__(self, main_window, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.main_window = main_window
		self.loading_packages = LoadingStatus(_("Loading Packages"), _("This should only take a moment"))
		self.uninstalling = LoadingStatus(_("Uninstalling Packages"), _("This should only take a moment"))
		self.uninstalling_view.set_content(self.uninstalling)
		self.reinstalling = LoadingStatus(_("Reinstalling Package"), _("This could take a while"), True, PackageInstallWorker.cancel)
		self.reinstalling_view.set_content(self.reinstalling)
		self.changing_version = LoadingStatus(_("Changing Version"), _("This could take a while"), True, ChangeVersionWorker.cancel)
		self.changing_version_view.set_content(self.changing_version)
		self.filter_settings = Gio.Settings.new("io.github.flattool.Warehouse.filter")
		self.is_result = False
		self.prev_status = None
		self.selected_rows = []
		self.current_row_for_properties = None
		self.on_backspace_handler = self.selection_uninstall

		# Apply
		self.loading_view.set_content(self.loading_packages)
		self.packages_list_box.set_filter_func(self.filter_func)
		self.packages_list_box.set_sort_func(self.sort_func)
		self.properties_page.packages_page = self
		self.filters_page.packages_page = self
		self.__class__.instance = self

		# Connections
		self.search_entry.connect("search-changed", self.on_invalidate)
		self.search_bar.set_key_capture_widget(main_window)
		self.packages_list_box.connect("row-activated", self.row_activate_handler)
		self.select_button.connect("toggled", self.select_button_handler)
		self.filter_button.connect("toggled", self.filter_button_handler)
		self.reset_filters_button.connect("clicked", lambda *_: self.filters_page.reset_filters())
		self.packages_split.connect("notify::show-content", self.filter_page_handler)
		self.packages_bpt.connect("apply", self.filter_page_handler)
		self.select_all_button.connect("clicked", self.select_all_handler)
		self.copy_menu.connect("row-activated", self.selection_copy)
		self.uninstall_button.connect("clicked", self.selection_uninstall)



from gi.repository import Gtk, GLib


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/gtk/loading_status.ui")
class LoadingStatus(Gtk.ScrolledWindow):
	__gtype_name__ = "LoadingStatus"
	gtc = Gtk.Template.Child

	title_label = gtc()
	description_label = gtc()
	progress_clamp = gtc()
	progress_bar = gtc()
	progress_label = gtc()
	button = gtc()

	def __init__(self, title, description, show_progress=False, on_cancel=None, **kwargs):
		super().__init__(**kwargs)

		self.title_label.set_label(GLib.markup_escape_text(title))
		self.description_label.set_label(GLib.markup_escape_text(description))
		self.progress_clamp.set_visible(show_progress)
		if on_cancel is None:
			self.button.set_visible(False)
		else:
			self.button.connect("clicked", lambda *_: on_cancel())



from gi.repository import Adw, Gtk, Gio
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
from src.gtk.loading_status import LoadingStatus
from src.gtk.installation_chooser import InstallationChooser
import subprocess, re
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/remotes_page/add_remote_dialog.ui")
class AddRemoteDialog(Adw.Dialog):
	__gtype_name__ = "AddRemoteDialog"
	gtc = Gtk.Template.Child

	toast_overlay = gtc()
	cancel_button = gtc()
	apply_button = gtc()
	content_page = gtc()
	title_row = gtc()
	name_row = gtc()
	url_row = gtc()
	installation_chooser = gtc()
	is_open = False

	def on_apply(self, *args):
		self.parent_page.status_stack.set_visible_child(self.parent_page.adding_view)
		self.apply_button.set_sensitive(False)
		error = [None]

		def thread(*args):
			HostInfo.main_window.add_refresh_lockout("adding remote")
			cmd = [
				"flatpak",
				"remote-add",
				f"--title={self.title_row.get_text()}",
				self.name_row.get_text(),
				self.url_row.get_text(),
			]
			installation = self.installation_chooser.get_installation()
			if installation == "user" or installation == "system":
				cmd.append(f"--{installation}")
			else:
				cmd.append(f"--installation={installation}")

			try:
				subprocess.run(cmd, check=True, capture_output=True, text=True)
			except subprocess.CalledProcessError as cpe:
				error[0] = cpe.stderr
			except Exception as e:
				error[0] = e

		def callback(*args):
			HostInfo.main_window.remove_refresh_lockout("adding remote")
			if error[0]:
				self.parent_page.status_stack.set_visible_child(self.parent_page.main_view)
				self.apply_button.set_sensitive(True)
				self.parent_page.toast_overlay.add_toast(ErrorToast(_("Could not add remote"), str(error[0])).toast)
			else:
				self.main_window.refresh_handler()
				self.parent_page.toast_overlay.add_toast(Adw.Toast(title=_("Added {}").format(self.title_row.get_text())))

		Gio.Task.new(None, None, callback).run_in_thread(thread)
		self.close()

	def check_entries(self, row):
		is_passing = re.match(self.rexes[row], row.get_text())
		if is_passing:
			row.remove_css_class("error")
		else:
			row.add_css_class("error")

		match row:
			case self.title_row:
				self.title_passes = bool(is_passing)
			case self.name_row:
				self.name_passes = bool(is_passing)
			case self.url_row:
				self.url_passes = bool(is_passing)

		self.apply_button.set_sensitive(self.title_passes and self.name_passes and self.url_passes)

	def present(self, *args, **kwargs):
		if self.__class__.is_open:
			return

		self.__class__.is_open = True
		super().present(*args, **kwargs)

	def on_close(self, *args):
		self.__class__.is_open = False

	def __init__(self, main_window, parent_page, remote_info=None, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.string_list = Gtk.StringList(strings=HostInfo.installations)
		self.main_window = main_window
		self.parent_page = parent_page

		self.rexes = {
			self.title_row: r"^(?=.*[A-Za-z0-9])[A-Za-z0-9._-]+( +[A-Za-z0-9._-]+)*$",
			self.name_row: r"^[a-zA-Z0-9\-._]+$",
			self.url_row: r"^[a-zA-Z0-9\-._~:/?#[\]@!$&\'()*+,;= ]+$",
		}
		self.title_passes = False
		self.name_passes = False
		self.url_passes = False

		# Apply
		self.installation_chooser.set_content_strings(_("remote"), False)
		if remote_info:
			self.title_row.set_text(remote_info["title"])
			self.name_row.set_text(remote_info["name"])
			self.url_row.set_text(remote_info["link"])
			if remote_info["description"] == "local file":
				self.check_entries(self.title_row)
				self.check_entries(self.name_row)
				self.check_entries(self.url_row)
				self.url_row.set_editable(False)
			else:
				self.title_row.set_editable(False)
				self.name_row.set_editable(False)
				self.url_row.set_editable(False)
				self.apply_button.set_sensitive(True)
		else:
			self.apply_button.set_sensitive(False)

		# Connections
		self.connect("closed", self.on_close)
		self.cancel_button.connect("clicked", lambda *_: self.close())
		self.apply_button.connect("clicked", self.on_apply)
		self.title_row.connect("changed", self.check_entries)
		self.name_row.connect("changed", self.check_entries)
		self.url_row.connect("changed", self.check_entries)



from gi.repository import Adw, Gtk
from src.host_info import HostInfo
from src.install_page.result_row import ResultRow
import gettext

_ = gettext.gettext


class AddedGroup(Adw.PreferencesGroup):
	__gtype_name__ = "AddedGroup"

	def add_row(self, row):
		self.rows.append(row)
		self.add(row)

	def rem_row(self, row):
		if row in self.rows:
			self.rows.remove(row)
			self.remove(row)

	def remove_all(self, *args):
		while len(self.rows) > 0 and (row := self.rows[0]):
			row.activate()

	def __init__(self, remote, installation, **kwargs):
		super().__init__(**kwargs)

		self.remote = remote
		self.installation = installation
		self.rows = []

		self.set_title(f"{remote.title}")
		self.set_description(_("Installation: {}").format(installation))

		remove_all = Gtk.Button(
			child=Adw.ButtonContent(
				icon_name="warehouse:list-remove-all-symbolic",
				label=_("Remove All"),
			),
			valign=Gtk.Align.CENTER,
		)
		remove_all.add_css_class("flat")
		remove_all.connect("clicked", self.remove_all)
		self.set_header_suffix(remove_all)


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/install_page/pending_page.ui")
class PendingPage(Adw.NavigationPage):
	__gtype_name__ = "PendingPage"
	gtc = Gtk.Template.Child

	stack = gtc()
	main_view = gtc()
	none_pending = gtc()
	preferences_page = gtc()
	install_button = gtc()

	def add_package_row(self, row):
		self.added_packages.append(row.package)
		row.set_state(ResultRow.PackageState.SELECTED)
		key = f"{row.package.remote}<>{row.package.installation}"
		added_row = ResultRow(row.package, ResultRow.PackageState.ADDED, row.origin_list_box)
		group = None
		try:
			group = self.groups[key]
			group.add_row(added_row)
		except KeyError:
			group = AddedGroup(added_row.package.remote, added_row.package.installation)
			group.add_row(added_row)
			self.groups[key] = group
			self.preferences_page.add(group)

		added_row.connect("activated", self.remove_package_row, group)
		self.stack.set_visible_child(self.main_view)

	def remove_package_row(self, row, group):
		# row.origin_row.set_state(ResultRow.PackageState.NEW)
		for item in row.origin_list_box:
			if item.state == ResultRow.PackageState.SELECTED and item.package.is_similar(row.package):
				item.set_state(ResultRow.PackageState.NEW)
				break

		group.rem_row(row)
		if row.package in self.added_packages:
			self.added_packages.remove(row.package)

		if len(group.rows) == 0:
			key = f"{row.package.remote}<>{row.package.installation}"
			self.groups.pop(key, None)
			self.preferences_page.remove(group)

		if len(self.added_packages) == 0:
			self.stack.set_visible_child(self.none_pending)

		install_page = HostInfo.main_window.pages[HostInfo.main_window.install_row]
		install_page.package_removed()

	def on_install(self, *args):
		package_requests = []
		for key, group in self.groups.items():
			item = {
				"remote": group.remote.name,
				"installation": group.installation,
				"package_names": [],
				"extra_flags": [],
			}
			for row in group.rows:
				item["package_names"].append(f"{row.package.app_id}//{row.package.branch}")

			package_requests.append(item)

		install_page = HostInfo.main_window.pages[HostInfo.main_window.install_row]
		install_page.install_packages(package_requests)

	def reset(self):
		for key, group in self.groups.items():
			self.preferences_page.remove(group)

		self.groups.clear()
		self.added_packages.clear()
		self.stack.set_visible_child(self.none_pending)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.groups = {}  # remote<>installation: adw.preference_group
		self.added_packages = []

		# Connections
		self.install_button.connect("clicked", self.on_install)

		# Apply



from gi.repository import GLib, Gio
from src.host_info import HostInfo
import subprocess, re
import gettext

_ = gettext.gettext


class ChangeVersionWorker:
	process = None
	callback = None
	error_callback = None
	loading_status = None
	did_error = False

	@classmethod
	def update_status(this, package_ratio, complete, total):
		final_ratio = (package_ratio + complete) / (total or 1)

		print(f"fr: {final_ratio:.2f}")
		print("r:", package_ratio, ", c:", complete, ", t:", total)
		print("=======================================")

		if not this.loading_status is None:
			GLib.idle_add(lambda *_: this.loading_status.progress_bar.set_fraction(final_ratio))

	@classmethod
	def change_version_thread(this, should_mask, package, commit):
		try:
			cmd = []

			installation = package.info["installation"]
			real_installation = ""
			if installation == "user":
				real_installation = f"--{installation}"
			elif installation == "system":
				cmd.append("pkexec")
				real_installation = f"--{installation}"
			else:
				real_installation = f"--installation={installation}"

			cmd += ["sh", "-c"]

			suffix = ""
			unmask_cmd = f"flatpak mask --remove {real_installation} {package.info['id']} && "
			change_version_cmd = f"flatpak update {package.info['ref']} --commit={commit} {real_installation} -y"
			mask_cmd = f" && flatpak mask {real_installation} {package.info['id']}"
			if package.is_masked:
				suffix += unmask_cmd

			suffix += change_version_cmd
			if should_mask:
				suffix += mask_cmd

			cmd.append(suffix)

			this.process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
			percent_pattern = r"\d{1,3}%"
			amount_pattern = r"(\d+)/(\d+)"
			for line in this.process.stdout:
				line = line.strip()
				percent_match = re.search(percent_pattern, line)
				if percent_match:
					ratio = int(percent_match.group()[0:-1]) / 100.0
					amount_match = re.search(amount_pattern, line)
					if amount_match:
						amount = amount_match.group().split("/")
						complete = int(amount[0]) - 1
						total = int(amount[1])
						this.update_status(ratio, complete, total)
					else:
						this.update_status(ratio, 0, 1)

			this.process.wait(timeout=10)
			if error := this.process.communicate()[1].strip():
				this.on_error(_("Error occurred while changing version"), error)

		except subprocess.TimeoutExpired:
			this.process.terminate()
			this.on_error(_("Error occurred while changing version"), _("Failed to exit cleanly"))

		except Exception as e:
			this.process.terminate()
			this.on_error(_("Error occurred while changing version"), str(e))

	@classmethod
	def cancel(this):
		if this.process is None:
			return

		try:
			this.process.terminate()
			this.process.wait(timeout=10)
		except Exception as e:
			this.on_error(_("Could not cancel version change"), str(e))

	@classmethod
	def on_done(this, *args):
		this.process = None
		HostInfo.main_window.remove_refresh_lockout("changing version")
		if not this.loading_status is None:
			this.loading_status.progress_bar.set_fraction(0.0)

		if not this.callback is None:
			this.callback(this.did_error)

	@classmethod
	def on_error(this, user_facing_label, error_message):
		this.did_error = True
		if not this.error_callback is None:
			this.error_callback(user_facing_label, error_message)

	@classmethod
	def change_version(this, should_mask, package, commit, loading_status=None, callback=None, error_callback=None):
		if not this.process is None:
			this.on_error(_("Could not change version"), _("Another package is changing version."))
			return False

		this.loading_status = loading_status
		this.callback = callback
		this.error_callback = error_callback
		HostInfo.main_window.add_refresh_lockout("changing version")
		Gio.Task.new(None, None, this.on_done).run_in_thread(lambda *_: this.change_version_thread(should_mask, package, commit))
		return True



from gi.repository import Adw, Gtk, GLib, Gio
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
from src.snapshot_page.tar_worker import TarWorker
import os, subprocess, json
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/snapshot_page/snapshot_box.ui")
class SnapshotBox(Gtk.Box):
	__gtype_name__ = "SnapshotBox"
	gtc = Gtk.Template.Child

	title = gtc()
	date = gtc()
	version = gtc()
	apply_button = gtc()
	rename_button = gtc()
	rename_menu = gtc()
	rename_entry = gtc()
	apply_rename = gtc()
	trash_button = gtc()

	def create_json(self):
		try:
			data = {
				"snapshot_version": 1,
				"name": "",
			}
			with open(self.json_path, "w") as file:
				json.dump(data, file, indent=4)
				return None

		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not write data"), str(e)).toast)

	def update_json(self, key, value):
		try:
			with open(self.json_path, "r+") as file:
				data = json.load(file)
				data[key] = value
				file.seek(0)
				json.dump(data, file, indent=4)
				file.truncate()

		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not write data"), str(e)).toast)

	def load_from_json(self):
		if not os.path.exists(self.json_path):
			self.create_json()

		try:
			with open(self.json_path, "r") as file:
				data = json.load(file)
				name = data["name"]
				if name != "":
					self.title.set_label(GLib.markup_escape_text(name))
				else:
					self.title.set_label(_("No Name Set"))

		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not write data"), str(e)).toast)

	def on_rename(self, widget):
		if not self.valid_checker():
			return

		self.update_json("name", self.rename_entry.get_text().strip())
		self.load_from_json()
		self.rename_menu.popdown()

	def valid_checker(self, *args):
		text = self.rename_entry.get_text().strip()
		valid = not ("/" in text or "\0" in text) and len(text) > 0
		self.apply_rename.set_sensitive(valid)
		if valid:
			self.rename_entry.remove_css_class("error")
		else:
			self.rename_entry.add_css_class("error")

		return valid

	def on_trash(self, button):
		error = [None]
		path = f"{self.snapshots_path}{self.folder}"
		if self.snapshot_page.is_trash_dialog_open:
			return

		def thread(*args):
			try:
				subprocess.run(["gio", "trash", path], capture_output=True, text=True, check=True)
			except subprocess.CalledProcessError as cpe:
				error[0] = cpe.stderr
			except Exception as e:
				error[0] = str(e)

		def callback(*args):
			if not error[0] is None:
				self.toast_overlay.add_toast(ErrorToast(_("Could not trash snapshot"), error[0]).toast)
				return

			self.parent_page.on_trash()
			self.toast_overlay.add_toast(Adw.Toast.new(_("Trashed snapshot")))

		def on_response(_, response):
			self.snapshot_page.is_trash_dialog_open = False
			if response != "continue":
				return

			Gio.Task.new(None, None, callback).run_in_thread(thread)

		self.snapshot_page.is_trash_dialog_open = True
		dialog = Adw.AlertDialog(heading=_("Trash Snapshot?"), body=_("This snapshot will be sent to the trash"))
		dialog.add_response("cancel", _("Cancel"))
		dialog.add_response("continue", _("Trash"))
		dialog.set_response_appearance("continue", Adw.ResponseAppearance.DESTRUCTIVE)
		dialog.connect("response", on_response)
		dialog.present(HostInfo.main_window)

	def get_fraction(self):
		loading_status = self.snapshot_page.snapshotting_status
		loading_status.progress_bar.set_fraction(self.worker.fraction)
		if self.worker.stop:
			self.snapshot_page.status_stack.set_visible_child(self.snapshot_page.split_view)
			self.parent_page.set_snapshots(self.parent_page.package_or_folder, True)
			properties_page = HostInfo.main_window.pages[HostInfo.main_window.packages_row].properties_page
			properties_page.set_properties(properties_page.package, True)
			data_page = HostInfo.main_window.pages[HostInfo.main_window.user_data_row]
			data_page.start_loading()
			data_page.end_loading()
			if self.worker in self.snapshot_page.workers:
				self.snapshot_page.workers.remove(self.worker)

			return False  # Stop the timeout
		else:
			return True  # Continue the timeout

	def on_apply(self, button):
		def on_response(dialog, response):
			if response != "continue":
				return

			self.snapshot_page.snapshotting_status.title_label.set_label(_("Applying Snapshot"))
			self.snapshot_page.snapshotting_status.progress_label.set_visible(False)
			self.snapshot_page.snapshotting_status.progress_bar.set_fraction(0.0)
			self.snapshot_page.status_stack.set_visible_child(self.snapshot_page.snapshotting_view)
			self.snapshot_page.workers.append(self.worker)
			self.worker.extract()
			GLib.timeout_add(200, self.get_fraction)

		has_data = os.path.exists(self.worker.new_path)
		dialog = Adw.AlertDialog(
			heading=_("Apply Snapshot?"),
			body=_("Any current user data for this app will be trashed") if has_data else "",
		)
		dialog.add_response("cancel", _("Cancel"))
		dialog.add_response("continue", _("Apply"))
		dialog.connect("response", on_response)
		dialog.present(HostInfo.main_window)

	def __init__(self, parent_page, folder, snapshots_path, toast_overlay, **kwargs):
		super().__init__(**kwargs)

		self.snapshot_page = parent_page.parent_page
		self.toast_overlay = toast_overlay
		self.app_id = snapshots_path.split("/")[-2].strip()
		self.worker = TarWorker(
			existing_path=f"{snapshots_path}{folder}",
			new_path=f"{HostInfo.home}/.var/app/{self.app_id}/",
			toast_overlay=self.toast_overlay,
		)

		split_folder = folder.split("_")
		if len(split_folder) < 2:
			return

		self.parent_page = parent_page
		self.folder = folder
		self.snapshots_path = snapshots_path
		self.epoch = int(split_folder[0])
		date_data = GLib.DateTime.new_from_unix_local(self.epoch).format("%x %X")
		self.date.set_label(date_data)
		self.version.set_label(_("Version: {}").format(split_folder[1].replace(".tar.zst", "")))
		self.json_path = f"{snapshots_path}{folder.replace('tar.zst', 'json')}"
		self.load_from_json()
		self.apply_button.connect("clicked", self.on_apply)
		self.apply_rename.connect("clicked", self.on_rename)
		self.rename_entry.connect("activate", self.on_rename)
		self.rename_entry.connect("changed", self.valid_checker)
		self.trash_button.connect("clicked", self.on_trash)



from gi.repository import GLib, Gio
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
import os, subprocess, json
import gettext

_ = gettext.gettext


class TarWorker:
	def compress_thread(self, *args):
		try:
			if not os.path.exists(self.new_path):
				os.makedirs(self.new_path)

			self.total = int(subprocess.run(["du", "-s", self.existing_path], check=True, text=True, capture_output=True).stdout.split("\t")[0])
			self.total /= 2.2  # estimate for space savings
			self.process = subprocess.Popen(
				["tar", "cafv", f"{self.new_path}/{self.file_name}.tar.zst", "-C", self.existing_path, "."], stdout=subprocess.PIPE, stderr=subprocess.PIPE
			)
			stdout, stderr = self.process.communicate()
			if self.process.returncode != 0:
				raise subprocess.CalledProcessError(self.process.returncode, self.process.args, output=stdout, stderr=stderr)

			with open(f"{self.new_path}/{self.file_name}.json", "w") as file:
				data = {
					"snapshot_version": 1,
					"name": self.name,
				}
				json.dump(data, file, indent=4)

			self.stop = True  # tell the check timeout to stop, because we know the file is done being made
			HostInfo.main_window.remove_refresh_lockout("managing snapshot")

		except subprocess.CalledProcessError as cpe:
			self.do_cancel(cpe.stderr.decode())  # stderr is in bytes, so decode it

		except Exception as e:
			self.do_cancel(str(e))

	def extract_thread(self, *args):
		try:
			if os.path.exists(self.new_path):
				subprocess.run(
					["gio", "trash", self.new_path], capture_output=True, check=True
				)  # trash the current user data, because new data will go in its place

			os.makedirs(self.new_path)  # create the new user data path

			self.total = int(subprocess.run(["du", "-s", self.existing_path], check=True, text=True, capture_output=True).stdout.split("\t")[0])
			self.total *= 2.2  # estimate from space savings
			self.process = subprocess.Popen(["tar", "--zstd", "-xvf", self.existing_path, "-C", self.new_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			stdout, stderr = self.process.communicate()
			if self.process.returncode != 0:
				raise subprocess.CalledProcessError(self.process.returncode, self.process.args, output=stdout, stderr=stderr)

			self.stop = True  # tell the check timeout to stop, because we know the file is done being made
			HostInfo.main_window.remove_refresh_lockout("managing snapshot")

		except subprocess.CalledProcessError as cpe:
			self.do_cancel(cpe.stderr.decode())

		except Exception as e:
			self.do_cancel(str(e))

	def do_cancel(self, error_str):
		if self.has_cancelled or self.stop:
			return

		self.has_cancelled = True
		self.process.terminate()
		self.process.wait()
		if len(self.files_to_trash_on_cancel) > 0:
			try:
				subprocess.run(["gio", "trash"] + self.files_to_trash_on_cancel, capture_output=True, check=True)

			except Exception:
				pass

		self.stop = True
		HostInfo.main_window.remove_refresh_lockout("managing snapshot")
		if self.toast_overlay and error_str != "manual_cancel":
			self.toast_overlay.add_toast(ErrorToast(_("Error in snapshot handling"), error_str).toast)

	def check_size(self, check_path):
		try:
			output = subprocess.run(["du", "-s", check_path], check=True, text=True, capture_output=True).stdout.split("\t")[0]
			working_total = float(output)
			self.fraction = working_total / self.total
			return not self.stop

		except subprocess.CalledProcessError:
			return not self.stop  # continue the timeout or stop the timeout

	def compress(self):
		self.stop = False
		self.files_to_trash_on_cancel = [f"{self.new_path}/{self.file_name}.tar.zst", f"{self.new_path}/{self.file_name}.json"]
		HostInfo.main_window.add_refresh_lockout("managing snapshot")
		Gio.Task.new(None, None, None).run_in_thread(self.compress_thread)
		GLib.timeout_add(200, self.check_size, f"{self.new_path}/{self.file_name}.tar.zst")

	def extract(self):
		self.stop = False
		self.files_to_trash_on_cancel = [self.new_path]
		HostInfo.main_window.add_refresh_lockout("managing snapshot")
		Gio.Task.new(None, None, None).run_in_thread(self.extract_thread)
		GLib.timeout_add(200, self.check_size, self.new_path)

	def __init__(self, existing_path, new_path, file_name="", name="", toast_overlay=None):
		self.existing_path = existing_path
		self.new_path = new_path
		self.file_name = file_name
		self.name = name
		self.should_check = False
		self.stop = False
		self.fraction = 0.0
		self.total = 0
		self.process = None
		self.toast_overlay = toast_overlay
		self.has_cancelled = False
		self.files_to_trash_on_cancel = []



from gi.repository import Adw, Gtk, GLib, Gio
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
import subprocess
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/remotes_page/remote_row.ui")
class RemoteRow(Adw.ActionRow):
	__gtype_name__ = "RemoteRow"
	gtc = Gtk.Template.Child

	suffix_label = gtc()
	filter_button = gtc()
	menu_pop = gtc()
	menu_listbox = gtc()

	copy_title = gtc()
	copy_name = gtc()
	enable_remote = gtc()
	disable_remote = gtc()
	remove = gtc()

	def enable_remote_handler(self, *args):
		if not self.remote.disabled:
			self.parent_page.toast_overlay.add_toast(ErrorToast(_("Could not enable remote"), _("Remote is already enabled")).toast)
			return

		has_error = []

		def thread(*args):
			cmd = ["flatpak", "remote-modify", "--enable", self.remote.name]
			if self.installation == "user" or self.installation == "system":
				cmd.append(f"--{self.installation}")
			else:
				cmd.append(f"--installation={self.installation}")

			try:
				subprocess.run(cmd, check=True, capture_output=True, text=True)
			except subprocess.CalledProcessError as cpe:
				has_error.append(str(cpe.stderr))
			except Exception as e:
				has_error.append(str(e))

		def callback(*args):
			if len(has_error) > 0:
				GLib.idle_add(lambda *args, cpe=cpe: self.parent_page.toast_overlay.add_toast(ErrorToast(_("Could not enable remote"), has_error[0]).toast))
				return

			self.remove_css_class("warning")
			self.set_icon_name("")
			self.set_tooltip_text("")
			self.remote.disabled = False
			self.parent_page.toast_overlay.add_toast(Adw.Toast(title=_("Enabled remote")))
			self.menu_listbox.get_row_at_index(2).set_visible(False)
			self.menu_listbox.get_row_at_index(3).set_visible(True)
			self.parent_page.total_disabled -= 1
			install_page = HostInfo.main_window.pages[HostInfo.main_window.install_row]
			install_page.start_loading()
			install_page.end_loading()
			filters_page = HostInfo.main_window.pages[HostInfo.main_window.packages_row].filters_page
			filters_page.generate_filters()
			if self.parent_page.total_disabled == 0:
				self.parent_page.show_disabled_button.set_active(False)
				self.parent_page.show_disabled_button.set_visible(False)

		Gio.Task.new(None, None, callback).run_in_thread(thread)

	def disable_remote_handler(self, *args):
		error = [None]

		def callback(*args):
			if error[0]:
				return

			self.add_css_class("warning")
			self.set_icon_name("warehouse:error-symbolic")
			self.set_tooltip_text(_("Remote is Disabled"))
			self.remote.disabled = True
			self.parent_page.toast_overlay.add_toast(Adw.Toast(title=_("Disabled remote")))
			self.menu_listbox.get_row_at_index(2).set_visible(True)
			self.menu_listbox.get_row_at_index(3).set_visible(False)
			self.set_visible(self.parent_page.show_disabled_button.get_active())
			self.parent_page.show_disabled_button.set_visible(True)
			self.parent_page.total_disabled += 1
			self.parent_page.none_visible_handler()
			install_page = HostInfo.main_window.pages[HostInfo.main_window.install_row]
			install_page.start_loading()
			install_page.end_loading()
			filters_page = HostInfo.main_window.pages[HostInfo.main_window.packages_row].filters_page
			filters_page.settings.reset("remotes-list")
			filters_page.all_remotes_switch.set_active(False)
			filters_page.generate_filters()
			filters_page.packages_page.apply_filters()

		def thread(*args):
			if self.remote.disabled:
				self.parent_page.toast_overlay.add_toast(ErrorToast(_("Could not disable remote"), _("Remote is already disabled")).toast)
				return

			cmd = ["flatpak", "remote-modify", "--disable", self.remote.name]
			if self.installation == "user" or self.installation == "system":
				cmd.append(f"--{self.installation}")
			else:
				cmd.append(f"--installation={self.installation}")

			try:
				subprocess.run(cmd, check=True, capture_output=True, text=True)
			except subprocess.CalledProcessError as cpe:
				GLib.idle_add(lambda *args, cpe=cpe: self.parent_page.toast_overlay.add_toast(ErrorToast(_("Could not disable remote"), str(cpe.stderr)).toast))
				error[0] = cpe
				return
			except Exception as e:
				GLib.idle_add(lambda *args, e=e: self.parent_page.toast_overlay.add_toast(ErrorToast(_("Could not disable remote"), str(e)).toast))
				error[0] = e
				return

		def on_response(_, response):
			if response != "continue":
				return

			Gio.Task.new(None, None, callback).run_in_thread(thread)

		dialog = Adw.AlertDialog(
			heading=_("Disable {}?").format(self.remote.title), body=_("Any installed apps from {} will stop receiving updates").format(self.remote.name)
		)
		dialog.add_response("cancel", _("Cancel"))
		dialog.add_response("continue", _("Disable"))
		dialog.set_response_appearance("continue", Adw.ResponseAppearance.DESTRUCTIVE)
		dialog.connect("response", on_response)
		dialog.present(self.parent_page.main_window)

	def on_menu_action(self, listbox, row):
		row = row.get_child()
		match row:
			case self.copy_title:
				HostInfo.clipboard.set(self.remote.title)
				self.parent_page.toast_overlay.add_toast(Adw.Toast(title=_("Copied title")))
			case self.copy_name:
				HostInfo.clipboard.set(self.remote.name)
				self.parent_page.toast_overlay.add_toast(Adw.Toast(title=_("Copied name")))
			case self.enable_remote:
				self.enable_remote_handler()
			case self.disable_remote:
				self.disable_remote_handler()
			case self.remove:
				self.parent_page.remove_remote(self)

		self.menu_pop.popdown()

	def idle_stuff(self):
		self.set_title(self.remote.title)
		self.set_subtitle(_("Installation: {}").format(self.installation))
		self.suffix_label.set_label(self.remote.name)
		if self.remote.disabled:
			self.set_icon_name("warehouse:error-symbolic")
			self.add_css_class("warning")
			self.set_tooltip_text(_("Remote is Disabled"))

	def __init__(self, parent_page, installation, remote, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.parent_page = parent_page
		self.remote = remote
		self.installation = installation

		# Apply
		GLib.idle_add(lambda *_: self.idle_stuff())

		## Show / Hide the Enable / Disable actions depending on remote status
		self.menu_listbox.get_row_at_index(2).set_visible(remote.disabled)
		self.menu_listbox.get_row_at_index(3).set_visible(not remote.disabled)

		# Connections
		self.menu_listbox.connect("row-activated", self.on_menu_action)
		self.filter_button.connect("clicked", lambda *_: parent_page.filter_remote(self))



from gi.repository import Adw, Gtk, GLib, Gio
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
from src.gtk.app_row import AppRow
from src.snapshot_page.snapshots_list_page import SnapshotsListPage
from src.gtk.sidebar_button import SidebarButton
from src.gtk.loading_status import LoadingStatus
from src.snapshot_page.new_snapshot_dialog import NewSnapshotDialog
from src.snapshot_page.tar_worker import TarWorker
from src.gtk.attempt_install_dialog import AttemptInstallDialog
import os, subprocess
import gettext

_ = gettext.gettext


class LeftoverSnapshotRow(Adw.ActionRow):
	__gtype_name__ = "LeftoverSnapshotRow"

	def idle_stuff(self):
		self.set_title(self.name)
		icon = Gtk.Image.new_from_icon_name("application-x-executable-symbolic")
		icon.set_icon_size(Gtk.IconSize.LARGE)
		self.add_prefix(icon)
		self.add_suffix(self.check_button)

	def gesture_handler(self, *args):
		self.on_long_press(self)

	def __init__(self, folder, on_long_press, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.folder = folder
		self.check_button = Gtk.CheckButton(visible=False)
		self.on_long_press = on_long_press
		self.rclick_gesture = Gtk.GestureClick(button=3)
		self.long_press_gesture = Gtk.GestureLongPress()

		# Apply
		self.add_controller(self.rclick_gesture)
		self.add_controller(self.long_press_gesture)
		self.check_button.add_css_class("selection-mode")
		self.name = self.folder.split(".")[-1]
		self.set_activatable(True)
		GLib.idle_add(lambda *_: self.idle_stuff())

		# Connections
		self.rclick_gesture.connect("released", self.gesture_handler)
		self.long_press_gesture.connect("pressed", self.gesture_handler)


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/snapshot_page/snapshot_page.ui")
class SnapshotPage(Adw.BreakpointBin):
	__gtype_name__ = "SnapshotPage"
	gtc = Gtk.Template.Child

	toast_overlay = gtc()
	sidebar_navpage = gtc()
	search_button = gtc()
	select_button = gtc()
	search_entry = gtc()
	search_bar = gtc()
	active_box = gtc()
	active_listbox = gtc()
	leftover_box = gtc()
	leftover_listbox = gtc()
	split_view = gtc()
	stack = gtc()
	no_snapshots = gtc()
	no_results = gtc()
	scrolled_window = gtc()
	status_open_button = gtc()
	status_new_button = gtc()
	new_button = gtc()
	status_stack = gtc()
	loading_view = gtc()
	snapshotting_view = gtc()
	select_all_button = gtc()
	copy_button = gtc()
	more_button = gtc()
	more_popover = gtc()
	more_menu = gtc()
	new_snapshots = gtc()
	apply_snapshots = gtc()
	install_from_snapshots = gtc()
	trash_snapshots = gtc()

	# Referred to in the main window
	# It is used to determine if a new page should be made or not
	# This must be set to the created object from within the class's __init__ method
	instance = None
	page_name = "snapshots"
	is_trash_dialog_open = False
	last_activated_row = None

	def sort_snapshots(self, *args):
		self.active_snapshot_paks.clear()
		self.leftover_snapshots.clear()
		bad_folders = []

		if not os.path.exists(HostInfo.snapshots_path):
			try:
				os.makedirs(HostInfo.snapshots_path)
			except Exception as e:
				self.toast_overlay.add_toast(ErrorToast(_("Could not load Snapshots"), str(e)).toast)
				return

		for folder in os.listdir(HostInfo.snapshots_path):
			if folder.count(".") < 2 or " " in folder:
				bad_folders.append(folder)
				continue

			has_tar = False
			for file in os.listdir(f"{HostInfo.snapshots_path}{folder}"):
				if file.endswith(".tar.zst"):
					has_tar = True
					break

			if not has_tar:
				bad_folders.append(folder)
				continue

			try:
				pak = HostInfo.id_to_flatpak[folder]
				self.active_snapshot_paks.append(pak)
			except KeyError:
				self.leftover_snapshots.append(folder)

		for folder in bad_folders:
			try:
				subprocess.run(["gio", "trash", f"{HostInfo.snapshots_path}{folder}"])
			except Exception:
				pass

	def long_press_handler(self, row):
		self.select_button.set_active(True)
		row.check_button.set_active(not row.check_button.get_active())

	def generate_active_list(self):
		for pak in self.active_snapshot_paks:
			row = AppRow(pak, self.long_press_handler)
			row.check_button.connect("toggled", lambda *_, _row=row: self.row_select_handler(_row))
			self.active_listbox.append(row)

		if len(self.active_snapshot_paks) > 0:
			self.active_box.set_visible(True)
		else:
			self.active_box.set_visible(False)

	def generate_leftover_list(self):
		for folder in self.leftover_snapshots:
			row = LeftoverSnapshotRow(folder, self.long_press_handler)
			row.check_button.connect("toggled", lambda *_, _row=row: self.row_select_handler(_row))
			self.leftover_listbox.append(row)

		if len(self.leftover_snapshots) > 0:
			self.leftover_box.set_visible(True)
			if len(self.active_snapshot_paks) == 0:
				self.stack.set_visible_child(self.scrolled_window)
		else:
			self.leftover_box.set_visible(False)

	def active_select_handler(self, listbox, row, should_show_content=True, refresh=False):
		if row.check_button.get_visible():
			row.check_button.set_active(not row.check_button.get_active())
			return

		self.last_activated_row = row
		self.leftover_listbox.select_row(None)
		self.list_page.set_snapshots(row.package, refresh)
		self.split_view.set_show_content(should_show_content)

	def leftover_select_handler(self, listbox, row, should_show_content=True, refresh=False):
		if row.check_button.get_visible():
			row.check_button.set_active(not row.check_button.get_active())
			return

		self.last_activated_row = row
		self.active_listbox.select_row(None)
		self.list_page.set_snapshots(row.folder, refresh)
		self.split_view.set_show_content(should_show_content)

	def select_first_row(self):
		if row := self.active_listbox.get_row_at_index(0):
			self.active_listbox.select_row(row)
			self.active_select_handler(None, row, False, True)
		elif row := self.leftover_listbox.get_row_at_index(0):
			self.leftover_listbox.select_row(row)
			self.leftover_select_handler(None, row, False, True)

	def show_snapshot(self, package):
		i = 0
		while row := self.active_listbox.get_row_at_index(i):
			i += 1
			if row.package is package:
				self.active_listbox.select_row(row)
				self.active_select_handler(None, row, True)
				self.toast_overlay.add_toast(Adw.Toast(title=_("Showing snapshots for {}").format(package.info["name"])))
				break
		else:
			dialog = NewSnapshotDialog(self, self.snapshotting_status, self.refresh, [package])
			toast = Adw.Toast(title=_("No snapshots for {}").format(package.info["name"]), button_label=_("New"))
			toast.connect("button-clicked", lambda *_: dialog.present(HostInfo.main_window))
			self.toast_overlay.add_toast(toast)

	def start_loading(self):
		self.last_activated_row = None
		self.search_button.set_active(False)
		self.workers.clear()
		self.select_button.set_active(False)
		self.status_stack.set_visible_child(self.loading_view)
		self.active_box.set_visible(True)
		self.active_listbox.remove_all()
		self.leftover_box.set_visible(True)
		self.leftover_listbox.remove_all()
		self.selected_active_rows.clear()
		self.selected_leftover_rows.clear()

	def end_loading(self):
		def callback(*args):
			self.generate_active_list()
			self.generate_leftover_list()
			if (not self.active_box.get_visible()) and (not self.leftover_box.get_visible()):
				GLib.idle_add(lambda *_: self.status_stack.set_visible_child(self.no_snapshots))
			else:
				self.select_first_row()
				GLib.idle_add(lambda *_: self.stack.set_visible_child(self.scrolled_window))
				GLib.idle_add(lambda *_: self.status_stack.set_visible_child(self.split_view))

			data_exists = False
			for package in HostInfo.flatpaks:
				if package.info["id"] == "io.github.flattool.Warehouse":
					continue

				if os.path.exists(package.data_path):
					data_exists = True
					break

			if data_exists:
				self.new_button.set_sensitive(True)
				self.new_button.set_tooltip_text(None)
				self.status_new_button.set_sensitive(True)
				self.status_new_button.set_tooltip_text(None)
			else:
				self.new_button.set_sensitive(False)
				self.new_button.set_tooltip_text(_("No Data Found to Snapshot"))
				self.status_new_button.set_sensitive(False)
				self.status_new_button.set_tooltip_text(_("No Data Found to Snapshot"))

		Gio.Task.new(None, None, callback).run_in_thread(self.sort_snapshots)

	def open_snapshots_folder(self, button):
		try:
			Gio.AppInfo.launch_default_for_uri(f"file://{HostInfo.snapshots_path}", None)
			self.toast_overlay.add_toast(Adw.Toast.new(_("Opened snapshots folder")))
		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not open folder"), str(e)).toast)

	def on_cancel(self):
		for worker in self.workers:
			worker.do_cancel("manual_cancel")

		if self.new_snapshot_dialog is None:
			return

		for worker in self.new_snapshot_dialog.workers:
			worker.do_cancel("manual_cancel")

	def on_new(self, *args):
		self.new_snapshot_dialog = NewSnapshotDialog(self, self.snapshotting_status, self.refresh)
		self.new_snapshot_dialog.present(HostInfo.main_window)

	def refresh(self):
		self.start_loading()
		self.end_loading()

	def on_search(self, search_entry):
		text = search_entry.get_text().lower()
		i = 0
		total_active_visible = 0
		while row := self.active_listbox.get_row_at_index(i):
			i += 1
			row.set_visible(False)
			if text in row.get_title().lower():
				row.set_visible(True)
				total_active_visible += 1
		self.active_box.set_visible(total_active_visible > 0)

		i = 0
		total_leftover_visible = 0
		while row := self.leftover_listbox.get_row_at_index(i):
			i += 1
			row.set_visible(False)
			if text in row.get_title().lower():
				row.set_visible(True)
				total_leftover_visible += 1
		self.leftover_box.set_visible(total_leftover_visible > 0)

		if total_active_visible > 0 or total_leftover_visible > 0:
			self.stack.set_visible_child(self.scrolled_window)
		else:
			self.stack.set_visible_child(self.no_results)

	def sort_func(self, row1, row2):
		if type(row1) is AppRow:
			return row1.package.info["name"].lower() > row2.package.info["name"].lower()
		else:
			return row1.name.lower() > row2.name.lower()

	def set_selection_mode(self, *args):
		enable = self.select_button.get_active()
		self.active_listbox.set_selection_mode(Gtk.SelectionMode.NONE if enable else Gtk.SelectionMode.SINGLE)
		self.leftover_listbox.set_selection_mode(Gtk.SelectionMode.NONE if enable else Gtk.SelectionMode.SINGLE)
		if self.last_activated_row in self.active_listbox:
			self.active_listbox.select_row(self.last_activated_row)
		elif self.last_activated_row in self.leftover_listbox:
			self.leftover_listbox.select_row(self.last_activated_row)

		i = 0
		while row := self.active_listbox.get_row_at_index(i):
			i += 1
			row.check_button.set_visible(enable)
			if not enable:
				row.check_button.set_active(False)

		i = 0
		while row := self.leftover_listbox.get_row_at_index(i):
			i += 1
			row.check_button.set_visible(enable)
			if not enable:
				row.check_button.set_active(False)

	def select_all_handler(self, *args):
		i = 0
		while row := self.active_listbox.get_row_at_index(i):
			i += 1
			row.check_button.set_active(True)

		i = 0
		while row := self.leftover_listbox.get_row_at_index(i):
			i += 1
			row.check_button.set_active(True)

	def row_select_handler(self, row):
		if type(row) is AppRow:
			if row.check_button.get_active():
				self.selected_active_rows.append(row)
			elif row in self.selected_active_rows:
				self.selected_active_rows.remove(row)
		elif type(row) is LeftoverSnapshotRow:
			if row.check_button.get_active():
				self.selected_leftover_rows.append(row)
			elif row in self.selected_leftover_rows:
				self.selected_leftover_rows.remove(row)

		total_active = len(self.selected_active_rows)
		total_leftover = len(self.selected_leftover_rows)
		total = total_active + total_leftover
		self.sidebar_navpage.set_title(_("{} Selected").format(total_active + total_leftover) if total > 0 else _("Snapshots"))
		self.new_snapshots.set_visible(total_active > 0)
		self.copy_button.set_sensitive(total > 0)
		self.more_button.set_sensitive(total > 0)
		i = 0
		while row := self.more_menu.get_row_at_index(i):
			i += 1
			match row.get_child():
				case self.new_snapshots:
					row.set_visible(total_active > 0 and total_leftover == 0)
				case self.apply_snapshots:
					row.set_visible(total_active > 0 and total_leftover == 0)
				case self.install_from_snapshots:
					row.set_visible(total_active == 0 and total_leftover > 0)

	def select_copy_handler(self, *args):
		to_copy = ""
		i = 0
		while row := self.active_listbox.get_row_at_index(i):
			i += 1
			if row.check_button.get_active():
				to_copy += f"{HostInfo.snapshots_path}{row.package.info['id']}\n"

		i = 0
		while row := self.leftover_listbox.get_row_at_index(i):
			i += 1
			if row.check_button.get_active():
				to_copy += f"{HostInfo.snapshots_path}{row.folder}\n"

		to_copy = to_copy[0:-1]
		HostInfo.clipboard.set(to_copy)
		self.toast_overlay.add_toast(Adw.Toast(title=_("Copied Snapshot Paths")))

	def select_new_handler(self):
		packages = []
		for row in self.selected_active_rows:
			if os.path.exists(row.package.data_path):
				packages.append(row.package)

		if len(packages) == 0:
			self.toast_overlay.add_toast(Adw.Toast(title=_("No apps in your selection can be snapshotted")))
			return

		self.new_snapshot_dialog = NewSnapshotDialog(self, self.snapshotting_status, self.refresh, packages)
		self.new_snapshot_dialog.present(HostInfo.main_window)

	def get_snapshots_from_entry(self, app_ids):
		id_to_tar = {}
		for app_id in app_ids:
			path = f"{HostInfo.snapshots_path}{app_id}"
			if not os.path.exists(path):
				continue

			tarlist = []
			for file in os.listdir(path):
				if file.endswith(".tar.zst"):
					tarlist.append(file)

			id_to_tar[app_id] = tarlist
			if len(tarlist) < 1:
				id_to_tar.pop(app_id, None)

		return id_to_tar

	def get_total_fraction(self):
		total = 0
		stopped_workers_amount = 0
		for worker in self.workers:
			total += worker.fraction
			if worker.stop:
				stopped_workers_amount += 1

			if stopped_workers_amount == len(self.workers):
				self.snapshotting_status.progress_bar.set_fraction(1)
				self.snapshotting_status.progress_label.set_label(f"{len(self.workers)} / {len(self.workers)}")
				HostInfo.main_window.refresh_handler()
				self.workers.clear()
				return False

		self.snapshotting_status.progress_label.set_label(f"{stopped_workers_amount + 1} / {len(self.workers)}")
		self.snapshotting_status.progress_bar.set_fraction(total / len(self.workers))
		return True

	def on_apply_response(self, dialog, response):
		if response != "continue":
			return

		app_ids = []
		for row in self.selected_active_rows:
			app_ids.append(row.package.info["id"])

		for row in self.selected_leftover_rows:
			app_ids.append(row.folder)

		id_to_tar = self.get_snapshots_from_entry(app_ids)
		for app_id in id_to_tar:
			biggest = 0
			for tar in id_to_tar[app_id]:
				epoch = int(tar.split("_")[0])
				if epoch > biggest:
					biggest = epoch

			id_to_tar[app_id] = tar

		for app_id, tar in id_to_tar.items():
			worker = TarWorker(
				existing_path=f"{HostInfo.snapshots_path}{app_id}/{tar}",
				new_path=f"{HostInfo.home}/.var/app/{app_id}/",
				toast_overlay=self.toast_overlay,
			)
			self.workers.append(worker)
			worker.extract()

		if len(self.workers) > 0:
			self.snapshotting_status.title_label.set_label(_("Applying Snapshots"))
			self.snapshotting_status.progress_bar.set_fraction(0.0)
			self.snapshotting_status.progress_label.set_visible(len(self.workers) > 1)
			self.status_stack.set_visible_child(self.snapshotting_view)
			GLib.timeout_add(200, self.get_total_fraction)
		else:
			self.toast_overlay.add_toast(ErrorToast(_("No snapshots to extract"), _("No snapshots were found to extract")))

	def select_apply_handler(self):
		dialog = Adw.AlertDialog(heading=_("Apply These Snapshots?"), body=_("This will trash the current apps' user data, and apply their newest snapshot"))
		dialog.add_response("cancel", _("Cancel"))
		dialog.add_response("continue", _("Continue"))
		dialog.connect("response", self.on_apply_response)
		dialog.present(HostInfo.main_window)

	def install_handler(self):
		package_names = []
		for row in self.selected_leftover_rows:
			package_names.append(row.folder)

		AttemptInstallDialog(package_names, lambda is_valid: self.select_button.set_active(not is_valid))

	def selection_trash_handler(self):
		if len(self.selected_active_rows) + len(self.selected_leftover_rows) < 1 or self.is_trash_dialog_open:
			return

		def on_response(dialog, response):
			self.is_trash_dialog_open = False
			to_trash = []
			if response != "continue":
				return

			for row in self.selected_active_rows:
				to_trash.append(f"{HostInfo.snapshots_path}{row.package.info['id']}")

			for row in self.selected_leftover_rows:
				to_trash.append(f"{HostInfo.snapshots_path}{row.folder}")

			try:
				subprocess.run(["gio", "trash"] + to_trash, check=True, text=True, capture_output=True)
				self.start_loading()
				self.end_loading()
				self.toast_overlay.add_toast(Adw.Toast(title=_("Trashed snapshots")))
			except subprocess.CalledProcessError as cpe:
				self.toast_overlay.add_toast(ErrorToast(_("Could not trash snapshots"), cpe.stderr).toast)

		self.is_trash_dialog_open = True
		dialog = Adw.AlertDialog(heading=_("Trash Snapshots?"), body=_("These apps' snapshots will be sent to the trash"))
		dialog.add_response("cancel", _("Cancel"))
		dialog.add_response("continue", _("Trash"))
		dialog.set_response_appearance("continue", Adw.ResponseAppearance.DESTRUCTIVE)
		dialog.connect("response", on_response)
		dialog.present(HostInfo.main_window)

	def more_menu_handler(self, listbox, row):
		self.more_popover.popdown()
		row = row.get_child()
		match row:
			case self.new_snapshots:
				self.select_new_handler()
			case self.apply_snapshots:
				self.select_apply_handler()
			case self.install_from_snapshots:
				self.install_handler()
			case self.trash_snapshots:
				self.selection_trash_handler()

	def __init__(self, main_window, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.__class__.instance = self
		self.main_window = main_window
		self.active_snapshot_paks = []
		self.selected_active_rows = []
		self.selected_leftover_rows = []
		self.workers = []
		self.leftover_snapshots = []
		self.list_page = SnapshotsListPage(self)
		self.snapshotting_status = LoadingStatus("Initial Title", _("This could take a while"), True, self.on_cancel)
		self.new_snapshot_dialog = None
		self.on_backspace_handler = self.selection_trash_handler
		self.on_escape_handler = lambda *_: self.select_button.set_active(False)

		# Apply
		self.search_bar.set_key_capture_widget(HostInfo.main_window)
		self.loading_view.set_content(LoadingStatus(_("Loading Snapshots"), _("This should only take a moment")))
		self.snapshotting_view.set_content(self.snapshotting_status)
		self.split_view.set_content(self.list_page)
		self.active_listbox.set_sort_func(self.sort_func)
		self.leftover_listbox.set_sort_func(self.sort_func)

		# Connections
		self.active_listbox.connect("row-activated", self.active_select_handler)
		self.leftover_listbox.connect("row-activated", self.leftover_select_handler)
		self.status_open_button.connect("clicked", self.open_snapshots_folder)
		self.status_new_button.connect("clicked", self.on_new)
		self.new_button.connect("clicked", self.on_new)
		self.search_entry.connect("search-changed", self.on_search)
		self.select_button.connect("toggled", self.set_selection_mode)
		self.select_all_button.connect("clicked", self.select_all_handler)
		self.copy_button.connect("clicked", self.select_copy_handler)
		self.more_menu.connect("row-activated", self.more_menu_handler)



from gi.repository import Adw, Gtk
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/gtk/attempt_install_dialog.ui")
class AttemptInstallDialog(Adw.AlertDialog):
	__gtype_name__ = "AttemptInstallDialog"
	gtc = Gtk.Template.Child

	preferences_group = gtc()

	def generate_list(self):
		for installation, remotes in HostInfo.remotes.items():
			for remote in remotes:
				if remote.disabled:
					continue

				row = Adw.ActionRow(title=remote.title, subtitle=_("Installation: {}").format(installation))
				row.remote_name = remote.name
				row.remote_installation = installation
				button = Gtk.CheckButton()
				row.add_prefix(button)
				row.check_button = button
				row.set_activatable_widget(button)
				self.rows.append(row)
				self.preferences_group.add(row)
				if len(self.rows) > 1:
					button.set_group(self.rows[0].check_button)
				else:
					button.activate()

	def on_response(self, dialog, response):
		if response != "continue":
			if not self.callback is None:
				self.callback(False)

			return

		active_row = None
		for row in self.rows:
			if row.check_button.get_active():
				active_row = row
				break

		if not active_row is None:
			install_page = HostInfo.main_window.pages[HostInfo.main_window.install_row]
			HostInfo.main_window.activate_row(HostInfo.main_window.install_row)
			self.callback(True)
			install_page.install_packages(
				[
					{
						"remote": row.remote_name,
						"installation": row.remote_installation,
						"package_names": self.package_names,
						"extra_flags": [],
					}
				]
			)
		elif not self.callback is None:
			self.callback(False)

	def __init__(self, package_names, callback=None, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.rows = []
		self.package_names = package_names
		self.callback = callback

		# Apply
		self.generate_list()
		if len(self.rows) == 1:
			self.set_extra_child(None)
		elif len(self.rows) < 1:
			HostInfo.main_window.toast_overlay.add_toast(ErrorToast(_("Can't find matching packages"), _("Your system has no remotes added")).toast)

		self.present(HostInfo.main_window)

		# Connections
		self.connect("response", self.on_response)



from gi.repository import Gtk, GLib, Gio
from src.host_info import HostInfo
from src.user_data_page.data_box import DataBox
from src.gtk.loading_status import LoadingStatus
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/user_data_page/data_subpage.ui")
class DataSubpage(Gtk.Stack):
	__gtype_name__ = "DataSubpage"
	gtc = Gtk.Template.Child

	scrolled_window = gtc()

	label_box = gtc()
	subtitle_size_box = gtc()
	title = gtc()
	subtitle = gtc()
	spinner = gtc()
	size_label = gtc()
	flow_box = gtc()

	# Statuses
	content_box = gtc()
	no_data = gtc()
	no_results = gtc()

	def human_readable_size(self):
		working_size = self.total_size
		units = ["KB", "MB", "GB", "TB"]
		# size *= 1024
		for unit in units:
			if working_size < 1024:
				return f"~ {round(working_size)} {unit}"
			working_size /= 1024
		return f"~ {round(working_size)} PB"

	def sort_func(self, box1, box2):
		import random

		# print(random.randint(1, 100), self.sort_mode, self.sort_ascend)
		i1 = None
		i2 = None
		if self.sort_mode == "name":
			i1 = box1.get_child().title.lower()
			i2 = box2.get_child().title.lower()

		if self.sort_mode == "id":
			i1 = box1.get_child().subtitle.lower()
			i2 = box2.get_child().subtitle.lower()

		if self.sort_mode == "size" and self.ready_to_sort_size:
			i1 = box1.get_child().size
			i2 = box2.get_child().size

		if i1 is None or i2 is None:
			return 0

		return i1 > i2 if self.sort_ascend else i1 < i2

	def box_size_callback(self, size):
		self.finished_boxes += 1
		self.total_size += size
		if self.finished_boxes == self.total_items:
			self.size_label.set_label(self.human_readable_size())
			self.spinner.set_visible(False)
			self.ready_to_sort_size = True
			if self.sort_mode == "size":
				self.flow_box.invalidate_sort()
				self.set_visible_child(self.content_box)
				GLib.idle_add(lambda *_: self.parent_page.status_stack.set_visible_child(self.parent_page.main_view))

	def trash_handler(self, trashed_box):
		self.flow_box.remove(trashed_box)
		if not self.flow_box.get_child_at_index(0):
			self.set_visible_child(self.no_data)
			self.parent_page.start_loading()
			self.parent_page.end_loading()

	def set_selection_mode(self, is_enabled):
		if not is_enabled:
			self.size_label.set_visible(True)
			self.subtitle.set_visible(False)

		idx = 0
		while box := self.flow_box.get_child_at_index(idx):
			idx += 1
			box = box.get_child()
			if not is_enabled:
				GLib.idle_add(lambda *_, box=box: box.check_button.set_active(False))

			GLib.idle_add(lambda *_, box=box: box.check_button.set_visible(is_enabled))
			GLib.idle_add(lambda *_, box=box: box.install_button.set_visible(box.is_leftover and not is_enabled))

		self.selected_boxes.clear()

	def box_select_handler(self, box):
		cb = box.check_button
		if cb.get_active():
			self.selected_boxes.append(box)
		else:
			try:
				self.selected_boxes.remove(box)
			except ValueError:
				pass

		total = len(self.selected_boxes)
		self.subtitle.set_visible(not total == 0)
		self.size_label.set_visible(total == 0)
		self.subtitle.set_label(_("{} Selected").format(total))
		self.parent_page.copy_button.set_sensitive(total)
		self.parent_page.trash_button.set_sensitive(total)
		self.parent_page.install_button.set_sensitive(total)
		self.parent_page.more_button.set_sensitive(total)

	def box_interact_handler(self, flow_box, box):
		box = box.get_child()
		cb = box.check_button
		if cb.get_visible():
			cb.set_active(not cb.get_active())

	def select_all_handler(self, *args):
		idx = 0
		while box := self.flow_box.get_child_at_index(idx):
			idx += 1
			box.get_child().check_button.set_active(True)

	def box_rclick_handler(self, box):
		if self.should_rclick:
			self.parent_page.select_button.set_active(True)
			box.check_button.set_active(not box.check_button.get_active())

	def generate_list(self, flatpaks, data):
		self.flow_box.remove_all()
		self.boxes.clear()
		self.ready_to_sort_size = False
		self.finished_boxes = 0
		self.total_size = 0
		self.total_items = len(data)
		self.parent_page.search_entry.set_editable(True)
		self.should_rclick = True
		if flatpaks:
			for i, pak in enumerate(flatpaks):
				box = DataBox(
					self,
					self.parent_page.toast_overlay,
					False,
					pak.info["name"],
					pak.info["id"],
					pak.data_path,
					pak.icon_path,
					self.box_size_callback,
					self.trash_handler,
				)
				box.check_button.connect("toggled", lambda *_, box=box: self.box_select_handler(box))
				self.boxes.append(box)
				self.flow_box.append(box)

		else:
			for i, folder in enumerate(data):
				box = DataBox(
					self,
					self.parent_page.toast_overlay,
					True,
					folder.split(".")[-1],
					folder,
					f"{HostInfo.home}/.var/app/{folder}",
					None,
					self.box_size_callback,
					self.trash_handler,
				)
				box.check_button.connect("toggled", lambda *_, box=box: self.box_select_handler(box))
				self.flow_box.append(box)

		idx = 0
		while box := self.flow_box.get_child_at_index(idx):
			idx += 1
			box.set_focusable(False)
			child = box.get_child()
			child.set_focusable(False)
			child.row.set_focusable(child.check_button.get_visible())
			rclick = Gtk.GestureClick(button=3)
			rclick.connect("released", lambda *_, child=child: self.box_rclick_handler(child))
			box.add_controller(rclick)
			long_press = Gtk.GestureLongPress()
			long_press.connect("pressed", lambda *_, child=child: self.box_rclick_handler(child))
			box.add_controller(long_press)

		if idx == 0:
			self.set_visible_child(self.no_data)
		elif self.sort_mode != "size":
			self.set_visible_child(self.content_box)
			self.parent_page.status_stack.set_visible_child(self.parent_page.main_view)

	def filter_func(self, box):
		search_text = self.parent_page.search_entry.get_text().lower()
		box = box.get_child()
		if search_text in box.title.lower() or search_text in box.subtitle.lower():
			self.is_result = True
			return True

	def on_invalidate(self, box):
		current_status = self.get_visible_child()
		if not current_status is self.no_results:
			self.prev_status = self.get_visible_child()

		self.is_result = False
		self.flow_box.invalidate_filter()
		if self.is_result:
			self.set_visible_child(self.prev_status)
		else:
			self.set_visible_child(self.no_results)

		if self.parent_page.search_entry.get_text().lower() != "" and self.total_items == 0:
			self.set_visible_child(self.no_results)
		elif self.total_items == 0:
			self.set_visible_child(self.no_data)

	def update_sort_mode(self):
		self.sort_ascend = self.settings.get_boolean("sort-ascend")
		self.sort_mode = self.settings.get_string("sort-mode")
		self.flow_box.invalidate_sort()

	def __init__(self, title, parent_page, is_active, main_window, **kwargs):
		super().__init__(**kwargs)

		GLib.idle_add(lambda *_: self.title.set_label(title))

		# self.select_button.connect("toggled", lambda *_: self.set_selection_mode(self.select_button.get_active()))
		# self.flow_box.connect("child-activated", lambda _, item: (cb := (row := item.get_child()).check_button).set_active((not cb.get_active()) if row.get_activatable() else False))

		# Extra Object Creation
		self.main_window = main_window
		self.parent_page = parent_page
		# self.is_active = is_active
		self.sort_mode = ""
		self.sort_ascend = False
		self.total_size = 0
		self.total_items = 0
		self.boxes = []
		self.selected_boxes = []
		self.ready_to_sort_size = False
		self.should_rclick = True
		self.finished_boxes = 0
		self.is_result = False
		self.prev_status = None
		self.settings = Gio.Settings.new("io.github.flattool.Warehouse.data_page")

		# Apply
		self.flow_box.set_sort_func(self.sort_func)
		self.flow_box.set_filter_func(self.filter_func)

		if is_active:
			self.no_data.set_icon_name("warehouse:error-symbolic")
			self.no_data.set_title(_("No Active Data"))
			self.no_data.set_description(_("Warehouse cannot see any active user data or your system has no active user data present"))
		else:
			self.no_data.set_icon_name("warehouse:check-plain-symbolic")
			self.no_data.set_title(_("No Leftover Data"))
			self.no_data.set_description(_("There is no leftover user data"))

		# Connections
		parent_page.search_entry.connect("search-changed", self.on_invalidate)
		self.flow_box.connect("child-activated", self.box_interact_handler)



# window.py
#
# Copyright 2023 Heliguy
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Adw, Gdk, Gio, Gtk
from src.host_info import HostInfo
from src.packages_page.packages_page import PackagesPage
from src.remotes_page.remotes_page import RemotesPage
from src.user_data_page.user_data_page import UserDataPage
from src.snapshot_page.snapshot_page import SnapshotPage
from src.install_page.install_page import InstallPage
from src.gtk.error_toast import ErrorToast
from src.const import Config
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/main_window/window.ui")
class WarehouseWindow(Adw.ApplicationWindow):
	__gtype_name__ = "WarehouseWindow"
	gtc = Gtk.Template.Child
	main_breakpoint = gtc()
	toast_overlay = gtc()
	file_drop_revealer = gtc()
	main_split = gtc()
	file_drop_view = gtc()
	stack = gtc()
	refresh_button = gtc()
	main_menu = gtc()
	navigation_row_listbox = gtc()
	packages_row = gtc()
	remotes_row = gtc()
	user_data_row = gtc()
	snapshots_row = gtc()
	install_row = gtc()

	def start_loading(self, *args):
		for _, page in self.pages.items():
			if page.instance:
				page.instance.start_loading()

	def end_loading(self, *args):
		for _, page in self.pages.items():
			if page.instance:
				page.instance.end_loading()

		self.refresh_button.set_sensitive(True)
		self.refresh_requested = False
		self.remove_refresh_lockout("refresh handler direct")

	def do_refresh(self):
		self.start_loading()
		self.refresh_button.set_sensitive(False)
		HostInfo.get_flatpaks(callback=self.end_loading)

	def refresh_handler(self, *args):
		if len(self.refresh_lockouts) == 0:
			self.add_refresh_lockout("refresh handler direct")
			self.do_refresh()
		elif "refresh handler direct" in self.refresh_lockouts:
			return
		else:
			self.refresh_requested = True

	def add_refresh_lockout(self, reason):
		self.refresh_lockouts.append(reason)
		self.refresh_button.set_sensitive(False)

	def remove_refresh_lockout(self, reason):
		if reason in self.refresh_lockouts:
			self.refresh_lockouts.remove(reason)

		if len(self.refresh_lockouts) == 0:
			if self.refresh_requested:
				self.do_refresh()
			else:
				self.refresh_button.set_sensitive(True)

	def navigation_handler(self, _, row):
		if self.main_split.get_collapsed():
			self.main_split.set_show_sidebar(False)

		row = row.get_child()
		page = self.pages[row]
		self.stack.set_visible_child(page)
		self.settings.set_string("page-shown", page.page_name)

	def activate_row(self, nav_row):
		idx = 0
		while row := self.navigation_row_listbox.get_row_at_index(idx):
			idx += 1
			if row.get_child() is nav_row:
				row.activate()
				break

	def show_saved_page(self):
		page_to_show = self.settings.get_string("page-shown")
		page_found = False
		for row, page in self.pages.items():
			self.stack.add_child(page)

			if page.page_name == page_to_show:
				page_found = True
				self.activate_row(row)

		if not page_found:
			self.navigation_row_listbox.get_row_at_index(0).activate()

	def on_file_drop(self, drop_target, value, x, y):
		# needed for X11
		self.on_drop_leave()

		try:
			paks = []
			remotes = []
			for file in value:
				path = file.get_path()
				if path.endswith(".flatpak") or path.endswith(".flatpakref"):
					paks.append(Gio.File.new_for_path(path))
				elif path.endswith(".flatpakrepo"):
					remotes.append(path)
				else:
					dialog = Adw.AlertDialog(
						heading=_("Unsupported Filetype"),
						body=_("Only .flatpak, .flatpakref, and .flatpakrepo files are supported."),
					)
					dialog.add_response("continue", _("OK"))
					dialog.present(self)
					return

			if len(remotes) > 0 and len(paks) > 0:
				dialog = Adw.AlertDialog(
					heading=_("Mixed Filetypes"),
					body=_("Flatpaks and remotes cannot be installed at the same time."),
				)
				dialog.add_css_class("error")
				dialog.add_response("continue", _("OK"))
				dialog.present(self)
				return

			if len(remotes) > 1:
				dialog = Adw.AlertDialog(
					heading=_("Too Many Remotes"),
					body=_("Only one remote at a time is supported."),
				)
				dialog.add_response("continue", _("OK"))
				dialog.present(self)
				return

			if len(remotes) == 1:
				# Adding a remote
				self.activate_row(self.remotes_row)
				remotes_page = self.pages[self.remotes_row]
				remotes_page.local_file_handler(remotes[0])
			elif len(paks) > 0:
				# Add packages
				self.activate_row(self.install_row)
				install_page = self.pages[self.install_row]
				install_page.select_page.file_dialog_handler(paks)

		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not open files"), str(e)).toast)

	def on_drop_enter(self, *args):
		self.main_split.add_css_class("blurred")
		self.file_drop_revealer.set_visible(True)
		self.file_drop_revealer.set_reveal_child(True)
		return 1

	def on_drop_leave(self, *args):
		self.main_split.remove_css_class("blurred")
		self.file_drop_revealer.set_reveal_child(False)
		self.file_drop_revealer.set_visible(False)

	def switch_page_shortcut_handler(self, letter):
		self.activate_row(self.shortcut_to_pages[letter])

	def key_handler(self, controller, keyval, keycode, state):
		page = self.stack.get_visible_child()
		if keyval == Gdk.KEY_BackSpace or keyval == Gdk.KEY_Delete:
			try:
				if page.select_button.get_active():
					page.on_backspace_handler()
			except AttributeError:
				pass
		elif keyval == Gdk.KEY_Escape:
			try:
				page.on_escape_handler()
			except AttributeError:
				pass

	def __init__(self, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		HostInfo.main_window = self
		ErrorToast.main_window = self
		self.settings = Gio.Settings.new("io.github.flattool.Warehouse")
		self.pages = {
			self.packages_row: PackagesPage(main_window=self),
			self.remotes_row: RemotesPage(main_window=self),
			self.user_data_row: UserDataPage(main_window=self),
			self.snapshots_row: SnapshotPage(main_window=self),
			self.install_row: InstallPage(main_window=self),
		}
		self.shortcut_to_pages = {
			"p": self.packages_row,
			"m": self.remotes_row,
			"d": self.user_data_row,
			"s": self.snapshots_row,
			"i": self.install_row,
		}
		self.navigation_row_listbox.connect("row-activated", self.navigation_handler)
		self.show_saved_page()
		self.refresh_lockouts = []
		self.refresh_requested = False
		file_drop = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
		event_controller = Gtk.EventControllerKey()

		# Apply
		self.add_controller(file_drop)
		self.add_controller(event_controller)
		self.settings.bind("window-width", self, "default-width", Gio.SettingsBindFlags.DEFAULT)
		self.settings.bind("window-height", self, "default-height", Gio.SettingsBindFlags.DEFAULT)
		self.settings.bind("is-maximized", self, "maximized", Gio.SettingsBindFlags.DEFAULT)
		self.settings.bind("is-fullscreen", self, "fullscreened", Gio.SettingsBindFlags.DEFAULT)
		if Config.DEVEL:
			self.add_css_class("devel")

		# Connections
		file_drop.connect("drop", self.on_file_drop)
		file_drop.connect("enter", self.on_drop_enter)
		file_drop.connect("leave", self.on_drop_leave)
		event_controller.connect("key-pressed", self.key_handler)
		self.refresh_button.connect("clicked", self.refresh_handler)

		# Apply again
		self.start_loading()
		HostInfo.get_flatpaks(callback=self.end_loading)



from gi.repository import Adw, Gtk, GLib, Gio
from src.gtk.error_toast import ErrorToast
from src.host_info import HostInfo
from src.gtk.loading_status import LoadingStatus
from src.change_version_page.change_version_worker import ChangeVersionWorker
import subprocess
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/change_version_page/change_version_page.ui")
class ChangeVersionPage(Adw.NavigationPage):
	__gtype_name__ = "ChangeVersionPage"
	gtc = Gtk.Template.Child
	toast_overlay = gtc()
	scrolled_window = gtc()
	versions_clamp = gtc()
	root_group_check_button = gtc()
	mask_group = gtc()
	mask_row = gtc()
	versions_group = gtc()
	action_bar = gtc()
	apply_button = gtc()

	selected_commit = None
	failure = None

	def get_commits(self, *args):
		cmd = ["sh", "-c"]
		script = f"LC_ALL=C flatpak remote-info --log {self.package.info['origin']} {self.package.info['ref']} "
		installation = self.package.info["installation"]
		if installation == "user" or installation == "system":
			script += f"--{installation}"
		else:
			script += f"--installation={installation}"

		cmd.append(script)

		commits = []
		changes = []
		dates = []
		try:
			output = subprocess.run(cmd, check=True, capture_output=True, text=True).stdout
			lines = output.strip().split("\n")
			for line in lines:
				line = line.strip().split(": ", 1)
				if len(line) < 2:
					continue
				elif line[0].startswith("Commit"):
					commits.append(line[1])
				elif line[0].startswith("Subject"):
					changes.append(line[1])
				elif line[0].startswith("Date"):
					dates.append(line[1])
		except subprocess.CalledProcessError as cpe:
			self.failure = cpe.stderr
			return
		except Exception as e:
			self.failure = str(e)
			return

		if not (len(commits) == len(changes) == len(dates)):
			self.failure = "Commits, Changes, and Dates are not of equivalent length"
			return

		def idle(*args):
			for index, commit in enumerate(commits):
				row = Adw.ActionRow(
					title=GLib.markup_escape_text(changes[index]), subtitle=f"{GLib.markup_escape_text(commit)}\n{GLib.markup_escape_text(dates[index])}"
				)
				if commit == self.package.cli_info.get("commit", None):
					row.set_sensitive(False)
					row.add_prefix(Gtk.Image(icon_name="warehouse:check-plain-symbolic", margin_start=5, margin_end=5))
					row.set_tooltip_text(_("Currently Installed Version"))
				else:
					check = Gtk.CheckButton()
					check.connect("activate", lambda *_, comm=commit: self.set_commit(comm))
					check.set_group(self.root_group_check_button)
					row.set_activatable_widget(check)
					row.add_prefix(check)

				self.versions_group.add(row)

		GLib.idle_add(idle)

	def set_commit(self, commit):
		self.selected_commit = commit

	def get_commits_callback(self, *args):
		if not self.failure is None:
			self.toast_overlay.add_toast(ErrorToast(_("Could not get versions"), self.failure).toast)
		else:
			self.scrolled_window.set_child(self.versions_clamp)

	def callback(self, did_error):
		HostInfo.main_window.refresh_handler()
		if not did_error:
			HostInfo.main_window.toast_overlay.add_toast(Adw.Toast(title=_("Changed {}'s Version").format(self.package.info["name"])))

	def error_callback(self, user_facing_label, error_message):
		HostInfo.main_window.toast_overlay.add_toast(ErrorToast(user_facing_label, error_message).toast)

	def on_apply(self, *args):
		if ChangeVersionWorker.change_version(
			self.mask_row.get_active(),
			self.package,
			self.selected_commit,
			self.packages_page.changing_version,
			self.callback,
			self.error_callback,
		):
			self.packages_page.set_status(self.packages_page.changing_version)

	def __init__(self, packages_page, package, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.packages_page = packages_page
		self.package = package

		# Apply
		pkg_name = package.info["name"]
		self.set_title(_("{} Versions").format(pkg_name))
		self.mask_row.set_subtitle(_("Ensure that {} will never be updated to a newer version").format(pkg_name))
		self.scrolled_window.set_child(LoadingStatus(_("Fetching Releases"), _("This could take a while")))
		Gio.Task.new(None, None, self.get_commits_callback).run_in_thread(self.get_commits)

		# Connections
		self.root_group_check_button.connect("toggled", lambda *_: self.action_bar.set_revealed(True))
		self.apply_button.connect("clicked", self.on_apply)



from gi.repository import Adw, Gtk, Gio
from src.host_info import HostInfo
from src.gtk.error_toast import ErrorToast
from src.snapshot_page.snapshot_box import SnapshotBox
from src.gtk.loading_status import LoadingStatus
from .new_snapshot_dialog import NewSnapshotDialog
import os
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/snapshot_page/snapshots_list_page.ui")
class SnapshotsListPage(Adw.NavigationPage):
	__gtype_name__ = "SnapshotsListPage"
	gtc = Gtk.Template.Child

	toolbar_view = gtc()
	listbox = gtc()
	toast_overlay = gtc()
	open_button = gtc()
	new_button = gtc()

	def thread(self, *args):
		is_leftover = type(self.package_or_folder) is str
		for snapshot in os.listdir(folder := f"{self.snapshots_path}{self.current_folder}/"):
			if snapshot.endswith(".json"):
				continue

			row = SnapshotBox(self, snapshot, folder, self.toast_overlay)
			row.apply_button.set_sensitive(not is_leftover)
			self.snapshots_rows.append(row)
			if is_leftover:
				row.apply_button.set_tooltip_text(_("App not Installed"))

	def callback(self, *args):
		if len(self.snapshots_rows) == 0:
			self.parent_page.refresh()
			return

		for i, row in enumerate(self.snapshots_rows):
			self.listbox.append(row)
			self.listbox.get_row_at_index(i).set_activatable(False)

	def set_snapshots(self, package_or_folder, refresh=False):
		if package_or_folder == self.package_or_folder and not refresh:
			return

		folder = None
		self.package_or_folder = package_or_folder
		if type(package_or_folder) is str:
			self.set_title(package_or_folder)
			folder = package_or_folder
			self.new_button.set_sensitive(False)
			self.new_button.set_tooltip_text(_("App not Installed"))
		else:
			folder = package_or_folder.info["id"]
			self.set_title(_("{} Snapshots").format(package_or_folder.info["name"]))
			if os.path.exists(package_or_folder.data_path):
				self.new_button.set_sensitive(True)
				self.new_button.set_tooltip_text(None)
			else:
				self.new_button.set_sensitive(False)
				self.new_button.set_tooltip_text(_("No Data Found to Snapshot"))

		self.current_folder = folder
		self.snapshots_rows.clear()
		self.listbox.remove_all()

		Gio.Task.new(None, None, self.callback).run_in_thread(self.thread)

	def open_snapshots_folder(self, button):
		path = f"{self.snapshots_path}{self.current_folder}/"
		try:
			if not os.path.exists(path):
				raise Exception(f"error: File '{path}' does not exist")

			Gio.AppInfo.launch_default_for_uri(f"file://{path}", None)
			self.toast_overlay.add_toast(Adw.Toast.new(_("Opened snapshots folder")))
		except Exception as e:
			self.toast_overlay.add_toast(ErrorToast(_("Could not open folder"), str(e)).toast)

	def on_done(self):
		self.parent_page.status_stack.set_visible_child(self.parent_page.split_view)
		self.set_snapshots(self.package_or_folder, refresh=True)

	def on_new(self, button):
		self.parent_page.new_snapshot_dialog = NewSnapshotDialog(self.parent_page, self.parent_page.snapshotting_status, self.on_done, [self.package_or_folder])
		self.parent_page.new_snapshot_dialog.present(HostInfo.main_window)

	def sort_func(self, row1, row2):
		row1 = row1.get_child()
		row2 = row2.get_child()
		return row1.epoch > row2.epoch

	def on_trash(self):
		self.set_snapshots(self.package_or_folder, refresh=True)

	def __init__(self, parent_page, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.parent_page = parent_page
		self.snapshots_path = HostInfo.snapshots_path
		self.current_folder = None
		self.package_or_folder = None
		self.snapshots_rows = []

		# Connections
		self.open_button.connect("clicked", self.open_snapshots_folder)
		self.new_button.connect("clicked", self.on_new)

		# Apply
		self.listbox.set_sort_func(self.sort_func)



# main.py
#
# Copyright 2023 Heliguy
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-only

import sys
import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")

from gi.repository import Gtk, Gio, Adw, GLib
from src.main_window.window import WarehouseWindow
from src.const import Config
from src.gtk.error_toast import ErrorToast
import gettext

_ = gettext.gettext


class WarehouseApplication(Adw.Application):
	"""The main application singleton class."""

	troubleshooting = "OS: {os}\nWarehouse version: {wv}\nGTK: {gtk}\nlibadwaita: {adw}\nApp ID: {app_id}\nProfile: {profile}\nLanguage: {lang}"
	version = Config.VERSION

	def __init__(self):
		super().__init__(
			application_id="io.github.flattool.Warehouse",
			flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE | Gio.ApplicationFlags.HANDLES_OPEN,
		)
		self.create_action("about", self.on_about_action)
		self.create_action("preferences", self.on_preferences_action)
		self.create_action("quit", lambda *_: self.quit(), ["<primary>q"])
		self.create_action("open-menu", lambda *_: self.props.active_window.main_menu.popup(), ["F10"])
		self.create_action("refresh", lambda *_: self.props.active_window.refresh_handler(), ["<primary>r", "F5"])
		self.create_action("open-files", self.on_open_files_shortcut, ["<primary>o"])

		self.create_action("show-packages-page", lambda *_: self.props.active_window.switch_page_shortcut_handler("p"), ["<primary>p"])
		self.create_action("show-remotes-page", lambda *_: self.props.active_window.switch_page_shortcut_handler("m"), ["<primary>m"])
		self.create_action("show-user-data-page", lambda *_: self.props.active_window.switch_page_shortcut_handler("d"), ["<primary>d"])
		self.create_action("show-snapshots-page", lambda *_: self.props.active_window.switch_page_shortcut_handler("s"), ["<primary>s"])
		self.create_action("show-install-page", lambda *_: self.props.active_window.switch_page_shortcut_handler("i"), ["<primary>i"])

		self.create_action("toggle-select-mode", self.on_toggle_select_mode_shortcut, ["<primary>b", "<primary>Return"])
		self.create_action("toggle-selection-kp-enter", self.on_toggle_select_mode_shortcut, ["<primary>KP_Enter"])  # Doesn't show in the shortcuts window
		self.create_action("search-mode", self.on_search_mode_shortcut, ["<primary>f"])
		self.create_action("filter", self.on_filter_shortcut, ["<primary>t"])
		self.create_action("new", self.on_new_shortcut, ["<primary>n"])
		self.create_action("active-data-view", lambda *_: self.on_data_view_shortcut(True), ["<Alt>1"])
		self.create_action("leftover-data-view", lambda *_: self.on_data_view_shortcut(False), ["<Alt>2"])

		self.is_dialog_open = False

		gtk_version = str(Gtk.MAJOR_VERSION) + "." + str(Gtk.MINOR_VERSION) + "." + str(Gtk.MICRO_VERSION)
		adw_version = str(Adw.MAJOR_VERSION) + "." + str(Adw.MINOR_VERSION) + "." + str(Adw.MICRO_VERSION)
		os_string = GLib.get_os_info("NAME") + " " + (GLib.get_os_info("VERSION") or "")
		lang = GLib.environ_getenv(GLib.get_environ(), "LANG")

		self.troubleshooting = self.troubleshooting.format(
			os=os_string,
			wv=self.version,
			gtk=gtk_version,
			adw=adw_version,
			profile=Config.PROFILE,
			app_id=self.get_application_id(),
			lang=lang,
		)

	def on_open_files_shortcut(self, *args):
		window = self.props.active_window

		def file_choose_callback(object, result):
			try:
				files = object.open_multiple_finish(result)
				if not files:
					window.toast_overlay.add_toast(ErrorToast(_("Could not add files"), _("No files were found")).toast)
					return

				window.on_file_drop(None, files, None, None)
			except GLib.GError as gle:
				if not (gle.domain == "gtk-dialog-error-quark" and gle.code == 2):
					window.toast_overlay.add_toast(ErrorToast(_("Could not add files"), str(gle)).toast)

		file_filter = Gtk.FileFilter(name=_("Flatpaks & Remotes"))
		file_filter.add_suffix("flatpak")
		file_filter.add_suffix("flatpakref")
		file_filter.add_suffix("flatpakrepo")
		filters = Gio.ListStore.new(Gtk.FileFilter)
		filters.append(file_filter)
		file_chooser = Gtk.FileDialog()
		file_chooser.set_filters(filters)
		file_chooser.set_default_filter(file_filter)
		file_chooser.open_multiple(window, None, file_choose_callback)

	def on_toggle_select_mode_shortcut(self, *args):
		try:
			button = self.props.active_window.stack.get_visible_child().select_button
			button.set_active(not button.get_active())
		except AttributeError:
			pass

	def on_search_mode_shortcut(self, *args):
		try:
			button = self.props.active_window.stack.get_visible_child().search_button
			button.set_active(True)
			entry = self.props.active_window.stack.get_visible_child().search_entry
			entry.grab_focus()
		except AttributeError:
			pass

	def on_filter_shortcut(self, *args):
		try:
			button = self.props.active_window.stack.get_visible_child().filter_button
			button.set_active(not button.get_active())
		except AttributeError:
			pass

		try:
			button = self.props.active_window.stack.get_visible_child().sort_button
			button.set_active(True)
		except AttributeError:
			pass

		try:
			button = self.props.active_window.stack.get_visible_child().show_disabled_button
			if button.get_visible():
				button.set_active(not button.get_active())
		except AttributeError:
			pass

	def on_new_shortcut(self, *args):
		page = self.props.active_window.stack.get_visible_child()
		try:
			page.new_custom_handler()
		except AttributeError:
			pass

		try:
			page.on_new()
		except AttributeError:
			pass

	def on_delete_shortcut(self, *args):
		page = self.props.active_window.stack.get_visible_child()
		try:
			if not page.select_button.get_active():
				return

			page.select_trash_handler()
		except AttributeError:
			pass

	def on_data_view_shortcut(self, is_active):
		page = self.props.active_window.stack.get_visible_child()
		try:
			adp = page.adp
			ldp = page.ldp
			page.stack.set_visible_child(adp if is_active else ldp)
		except AttributeError:
			pass

	def do_activate(self):
		"""Called when the application is activated.

		We raise the application's main window, creating it if
		necessary.
		"""
		win = self.props.active_window
		if not win:
			win = WarehouseWindow(application=self)
		win.present()

	def do_open(self, files, _n_files, _hint):
		"""Handle files opened from file manager or command line"""
		win = self.props.active_window
		if not win:
			win = WarehouseWindow(application=self)
		win.present()

		# Convert GFiles to file paths
		file_list = []
		for gfile in files:
			file_list.append(gfile)

		# Use existing file drop handler
		win.on_file_drop(None, file_list, None, None)

	def do_command_line(self, command_line):
		"""Handle command line arguments"""
		args = command_line.get_arguments()

		if len(args) > 1:  # First arg is program name
			files = []
			for path in args[1:]:
				gfile = Gio.File.new_for_path(path)
				if gfile.query_exists():
					files.append(gfile)

			if files:
				self.open(files, "")
				return 0

		# No valid files, just activate normally
		self.activate()
		return 0

	def on_about_action(self, widget, _a):
		"""Callback for the app.about action."""
		about = Adw.AboutDialog(
			application_name="Warehouse",
			application_icon="io.github.flattool.Warehouse",
			developer_name="Heliguy",
			version=self.version,
			developers=[
				"Heliguy https://github.com/heliguy4599",
				"kramo https://kramo.page",
			],
			artists=[
				"Heliguy https://github.com/heliguy4599",
				"kramo https://kramo.page",
				"Amy https://github.com/AtiusAmy",
				"eryn https://github.com/hericiumvevo",
			],
			copyright="© 2023 - 2025 Mitchell Winkelman",
			license_type=Gtk.License.GPL_3_0_ONLY,
			debug_info=self.troubleshooting,
			# Translators: do one of the following, one per line: Your Name, Your Name <email@email.org>, Your Name https://your.website
			translator_credits=_("translator-credits"),
			debug_info_filename="{}.txt".format(self.get_application_id()),
			website="https://github.com/flattool/warehouse",
			support_url="https://matrix.to/#/#warehouse-development:matrix.org",
			issue_url="https://github.com/flattool/warehouse/issues",
		)
		about.add_link(_("Donate"), "https://ko-fi.com/heliguy")
		about.add_credit_section(
			_("Contributors"),
			[
				# Contributors: do one of the following, one per line: Your Name, Your Name <email@email.org>, Your Name https://your.website
				"Win <winsdominoes2018@gmail.com>",
				"Óscar Fernández Díaz",
				"Runar https://github.com/runarcn",
				"Sunniva Løvstad <warehouse@turtle.garden>",
				"Maxim Therrien <maxim@mtsd.ca>",
			],
		)
		about.present(self.props.active_window)

	def on_preferences_action(self, widget, _):
		"""Callback for the app.preferences action."""
		print("app.preferences action activated")

	def create_action(self, name, callback, shortcuts=None):
		"""Add an application action.

		Args:
			name: the name of the action
			callback: the function to be called when the action is activated
			shortcuts: an optional list of accelerators
		"""
		action = Gio.SimpleAction.new(name, None)
		action.connect("activate", callback)
		self.add_action(action)
		if shortcuts:
			self.set_accels_for_action(f"app.{name}", shortcuts)


def main(version):
	"""The application's entry point."""
	app = WarehouseApplication()
	return app.run(sys.argv)



from gi.repository import Adw, Gtk, GLib
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/packages_page/uninstall_dialog.ui")
class UninstallDialog(Adw.AlertDialog):
	__gtype_name__ = "UninstallDialog"
	gtc = Gtk.Template.Child

	group = gtc()
	trash = gtc()
	is_open = False

	def on_response(self, dialog, response):
		self.__class__.is_open = False
		if response != "continue":
			return

		self.continue_callback(self.trash.get_active())

	def present(self, *args, **kwargs):
		if self.__class__.is_open:
			return

		self.__class__.is_open = True
		super().present(*args, **kwargs)

	def __init__(self, continue_callback, show_trash_option, package_name=None, **kwargs):
		super().__init__(**kwargs)

		if package_name:
			self.set_heading(GLib.markup_escape_text(_("Uninstall {}?").format(package_name)))
			self.set_body(GLib.markup_escape_text(_("It will not be possible to use {} after removal").format(package_name)))
		else:
			self.set_heading(GLib.markup_escape_text(_("Uninstall Packages?")))
			self.set_body(GLib.markup_escape_text(_("It will not be possible to use these packages after removal")))

		self.continue_callback = continue_callback
		self.add_response("cancel", _("Cancel"))
		self.add_response("continue", _("Uninstall"))
		self.set_response_appearance("continue", Adw.ResponseAppearance.DESTRUCTIVE)
		self.connect("response", self.on_response)
		self.group.set_title(GLib.markup_escape_text(_("App Settings & Content")))
		self.group.set_visible(show_trash_option)



from gi.repository import Adw, Gtk, GLib
from enum import Enum
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/install_page/result_row.ui")
class ResultRow(Adw.ActionRow):
	__gtype_name__ = "ResultRow"
	gtc = Gtk.Template.Child

	version_label = gtc()
	branch_label = gtc()
	add_image = gtc()
	sub_image = gtc()
	selected_image = gtc()
	installed_image = gtc()

	class PackageState(Enum):
		NEW = 0
		SELECTED = 1
		ADDED = 2
		INSTALLED = 3

	def idle_stuff(self):
		self.set_title(GLib.markup_escape_text(self.package.name))
		self.set_subtitle(self.package.app_id)
		self.version_label.set_label(GLib.markup_escape_text(self.package.version))
		self.branch_label.set_label(GLib.markup_escape_text(self.package.branch))
		self.version_label.set_visible(len(self.version_label.get_label()) != 0)
		self.branch_label.set_visible(len(self.branch_label.get_label()) != 0)

	def set_state(self, state):
		if state == self.state:
			return

		self.state = state
		self.add_image.set_visible(False)
		self.sub_image.set_visible(False)
		self.selected_image.set_visible(False)
		self.installed_image.set_visible(False)
		match state:
			case self.PackageState.NEW:
				self.set_sensitive(True)
				self.set_tooltip_text(_("Add Package to Queue"))
				self.add_image.set_visible(True)
			case self.PackageState.SELECTED:
				self.set_sensitive(False)
				self.set_tooltip_text(_("Package has been Added to Queue"))
				self.selected_image.set_visible(True)
			case self.PackageState.ADDED:
				self.set_sensitive(True)
				self.set_tooltip_text(_("Remove Package from Queue"))
				self.sub_image.set_visible(True)
			case self.PackageState.INSTALLED:
				self.set_sensitive(False)
				self.set_tooltip_text(_("This Package is Already Installed"))
				self.installed_image.set_visible(True)

	def __init__(self, package, package_state, origin_list_box, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.state = None
		self.package = package
		self.origin_list_box = origin_list_box

		# Connections

		# Apply
		GLib.idle_add(self.idle_stuff)
		self.set_state(package_state)



from gi.repository import Adw, Gtk, Gio
from src.gtk.error_toast import ErrorToast
from src.user_data_page.data_subpage import DataSubpage
from src.host_info import HostInfo
from src.gtk.sidebar_button import SidebarButton
from src.gtk.loading_status import LoadingStatus
from src.gtk.attempt_install_dialog import AttemptInstallDialog
import os, subprocess
import gettext

_ = gettext.gettext


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/user_data_page/user_data_page.ui")
class UserDataPage(Adw.BreakpointBin):
	__gtype_name__ = "UserDataPage"
	gtc = Gtk.Template.Child

	bpt = gtc()
	status_stack = gtc()
	loading_view = gtc()
	main_view = gtc()
	header_bar = gtc()
	switcher_bar = gtc()
	search_button = gtc()
	select_button = gtc()
	sort_button = gtc()
	search_bar = gtc()
	search_entry = gtc()
	toast_overlay = gtc()
	stack = gtc()
	revealer = gtc()

	sort_ascend = gtc()
	sort_descend = gtc()
	sort_name = gtc()
	sort_id = gtc()
	sort_size = gtc()

	select_all_button = gtc()
	copy_button = gtc()
	trash_button = gtc()
	install_button = gtc()
	more_button = gtc()
	more_popover = gtc()
	more_menu = gtc()
	more_trash = gtc()
	more_install = gtc()

	# Referred to in the main window
	# It is used to determine if a new page should be made or not
	# This must be set to the created object from within the class's __init__ method
	instance = None
	page_name = "user-data"
	data_path = f"{HostInfo.home}/.var/app"
	bpt_is_applied = False
	is_trash_dialog_open = False

	def sort_data(self, *args):
		self.data_flatpaks.clear()
		self.active_data.clear()
		self.leftover_data.clear()
		# paks = dict(HostInfo.id_to_flatpak)

		if not os.path.exists(self.data_path):
			return

		for folder in os.listdir(self.data_path):
			try:
				self.data_flatpaks.append(HostInfo.id_to_flatpak[folder])
				self.active_data.append(folder)
			except KeyError:
				self.leftover_data.append(folder)

	def start_loading(self, *args):
		self.status_stack.set_visible_child(self.loading_view)
		self.search_button.set_active(False)
		self.select_button.set_active(False)
		self.adp.size_label.set_label(_("Loading Size"))
		self.adp.spinner.set_visible(True)
		self.ldp.size_label.set_label(_("Loading Size"))
		self.ldp.spinner.set_visible(True)

	def end_loading(self, *args):
		def callback(*args):
			self.adp.generate_list(self.data_flatpaks, self.active_data)
			self.ldp.generate_list([], self.leftover_data)

		Gio.Task.new(None, None, callback).run_in_thread(self.sort_data)

	def sort_button_handler(self, button):
		if button in {self.sort_ascend, self.sort_descend}:
			self.settings.set_boolean("sort-ascend", self.sort_ascend.get_active())
		else:
			self.settings.set_string("sort-mode", self.buttons_to_sort_modes[button])

		self.adp.update_sort_mode()
		self.ldp.update_sort_mode()

	def load_sort_settings(self):
		mode = self.settings.get_string("sort-mode")
		ascend = self.settings.get_boolean("sort-ascend")
		self.sort_modes_to_buttons[mode].set_active(True)
		(self.sort_ascend if ascend else self.sort_descend).set_active(True)
		self.adp.update_sort_mode()
		self.ldp.update_sort_mode()

	def view_change_handler(self, *args):
		child = self.stack.get_visible_child()
		if child.total_size == 0:
			self.search_button.set_active(False)
			self.search_button.set_sensitive(False)
			self.select_button.set_active(False)
			self.select_button.set_sensitive(False)
			self.sort_button.set_active(False)
			self.sort_button.set_sensitive(False)
			self.search_entry.set_editable(False)
		else:
			self.search_button.set_sensitive(True)
			self.select_button.set_sensitive(True)
			self.sort_button.set_sensitive(True)
			self.search_entry.set_editable(True)

		self.more_button.set_visible(child is self.ldp and self.bpt_is_applied)
		self.install_button.set_visible(child is self.ldp and not self.bpt_is_applied)
		self.trash_button.set_visible(child is self.adp or not self.bpt_is_applied)

		has_selected = len(child.selected_boxes) > 0
		self.copy_button.set_sensitive(has_selected)
		self.trash_button.set_sensitive(has_selected)
		self.install_button.set_sensitive(has_selected)
		self.more_button.set_sensitive(has_selected)

	def select_toggle_handler(self, *args):
		active = self.select_button.get_active()
		self.adp.set_selection_mode(active)
		self.ldp.set_selection_mode(active)
		if not active:
			self.copy_button.set_sensitive(False)
			self.trash_button.set_sensitive(False)
			self.install_button.set_sensitive(False)
			self.more_button.set_sensitive(False)

	def select_all_handler(self, *args):
		child = self.stack.get_visible_child()
		child.select_all_handler()

	def copy_handler(self, *args):
		child = self.stack.get_visible_child()
		to_copy = ""
		for box in child.selected_boxes:
			to_copy += "\n" + box.data_path

		if len(to_copy) == 0:
			self.toast_overlay.add_toast(ErrorToast(_("Could not copy paths"), _("No boxes were selected")).toast)
		else:
			HostInfo.clipboard.set(to_copy.replace("\n", "", 1))
			self.toast_overlay.add_toast(Adw.Toast(title=_("Copied paths")))

	def selection_trash_handler(self, *args):
		error = [None]
		child = self.stack.get_visible_child()

		def thread(path):
			cmd = ["gio", "trash"] + path
			try:
				subprocess.run(cmd, check=True, capture_output=True, text=True)
				properties_page = HostInfo.main_window.pages[HostInfo.main_window.packages_row].properties_page
				properties_package = properties_page.package
				if not properties_package is None:
					properties_page.set_properties(properties_package, True)

				snapshot_list_page = HostInfo.main_window.pages[HostInfo.main_window.snapshots_row].list_page
				snapshot_list_package = snapshot_list_page.package_or_folder
				if not snapshot_list_package is None:
					snapshot_list_page.set_snapshots(snapshot_list_package, True)

			except subprocess.CalledProcessError as cpe:
				error[0] = cpe.stderr
			except Exception as e:
				error[0] = e

		def callback(*args):
			self.start_loading()
			self.end_loading()
			if error[0]:
				self.toast_overlay.add_toast(ErrorToast(_("Could not trash data"), str(error[0])).toast)
			else:
				self.toast_overlay.add_toast(Adw.Toast(title=_("Trashed data")))

		def on_response(dialog, response):
			self.is_trash_dialog_open = False
			if response != "continue":
				return

			to_trash = []
			for box in child.selected_boxes:
				to_trash.append(box.data_path)

			if len(to_trash) == 0:
				self.toast_overlay.add_toast(ErrorToast(_("Could not trash data"), _("No boxes were selected")).toast)
				return

			self.select_button.set_active(False)
			self.status_stack.set_visible_child(self.loading_view)
			Gio.Task.new(None, None, callback).run_in_thread(lambda *_: thread(to_trash))

		if len(child.selected_boxes) < 1 or self.is_trash_dialog_open:
			return

		self.is_trash_dialog_open = True
		dialog = Adw.AlertDialog(heading=_("Trash Data?"), body=_("Data will be sent to the trash"))
		dialog.add_response("cancel", _("Cancel"))
		dialog.add_response("continue", _("Continue"))
		dialog.set_response_appearance("continue", Adw.ResponseAppearance.DESTRUCTIVE)
		dialog.connect("response", on_response)
		dialog.present(ErrorToast.main_window)

	def breakpoint_handler(self, bpt, is_applied):
		self.bpt_is_applied = is_applied
		self.adp.label_box.set_orientation(Gtk.Orientation.VERTICAL if is_applied else Gtk.Orientation.HORIZONTAL)
		self.ldp.label_box.set_orientation(Gtk.Orientation.VERTICAL if is_applied else Gtk.Orientation.HORIZONTAL)
		child = self.stack.get_visible_child()
		self.install_button.set_visible(child is self.ldp and not is_applied)
		self.more_button.set_visible(child is self.ldp and is_applied)
		self.trash_button.set_visible(child is self.adp or not is_applied)

	def install_handler(self, *args):
		child = self.stack.get_visible_child()
		package_names = []
		for box in child.selected_boxes:
			package_names.append(box.subtitle)

		AttemptInstallDialog(package_names, lambda is_valid: self.select_button.set_active(not is_valid))

	def more_menu_handler(self, listbox, row):
		self.more_popover.popdown()
		row = row.get_child()
		match row:
			case self.more_install:
				self.install_handler()
			case self.more_trash:
				self.selection_trash_handler()

	def __init__(self, main_window, **kwargs):
		super().__init__(**kwargs)

		# Extra Object Creation
		self.__class__.instance = self
		self.adp = DataSubpage(_("Active Data"), self, True, main_window)
		self.ldp = DataSubpage(_("Leftover Data"), self, False, main_window)
		self.data_flatpaks = []
		self.active_data = []
		self.leftover_data = []
		self.total_items = 0
		self.settings = Gio.Settings.new("io.github.flattool.Warehouse.data_page")
		self.sort_modes_to_buttons = {
			"name": self.sort_name,
			"id": self.sort_id,
			"size": self.sort_size,
		}
		self.buttons_to_sort_modes = {}
		self.on_backspace_handler = self.selection_trash_handler
		self.on_escape_handler = lambda *_: self.select_button.set_active(False)

		# Apply
		for key, button in self.sort_modes_to_buttons.items():
			self.buttons_to_sort_modes[button] = key

		self.stack.add_titled_with_icon(
			child=self.adp,
			name="active",
			title=_("Active Data"),
			icon_name="warehouse:file-manager-symbolic",
		)
		self.stack.add_titled_with_icon(
			child=self.ldp,
			name="leftover",
			title=_("Leftover Data"),
			icon_name="warehouse:folder-templates-symbolic",
		)

		# Connections
		self.stack.connect("notify::visible-child", self.view_change_handler)
		self.select_button.connect("toggled", self.select_toggle_handler)
		self.select_all_button.connect("clicked", self.select_all_handler)
		self.copy_button.connect("clicked", self.copy_handler)
		self.trash_button.connect("clicked", self.selection_trash_handler)
		self.install_button.connect("clicked", self.install_handler)
		self.more_menu.connect("row-activated", self.more_menu_handler)
		self.sort_ascend.connect("clicked", self.sort_button_handler)
		self.sort_descend.connect("clicked", self.sort_button_handler)
		self.sort_name.connect("clicked", self.sort_button_handler)
		self.sort_id.connect("clicked", self.sort_button_handler)
		self.sort_size.connect("clicked", self.sort_button_handler)
		self.bpt.connect("apply", self.breakpoint_handler, True)
		self.bpt.connect("unapply", self.breakpoint_handler, False)

		# Apply again
		self.loading_view.set_content(LoadingStatus(_("Loading User Data"), _("This should only take a moment")))
		self.search_bar.set_key_capture_widget(main_window)
		self.load_sort_settings()



from gi.repository import Adw, Gtk, Gdk, GLib
import gettext

_ = gettext.gettext


class ErrorToast:
	main_window = None

	def __init__(self, display_msg, error_msg):
		def on_response(dialog, response_id):
			if response_id == "copy":
				self.clipboard.set(error_msg)

		# Extra Object Creation
		self.toast = Adw.Toast(title=display_msg, button_label=_("Details"))
		popup = Adw.AlertDialog.new(display_msg)
		self.clipboard = Gdk.Display.get_default().get_clipboard()

		# Apply
		print(display_msg)
		print(error_msg)
		popup.add_response("copy", _("Copy"))
		popup.add_response("ok", _("OK"))
		lb = Gtk.Label(selectable=True, wrap=True)  # , natural_wrap_mode=Gtk.NaturalWrapMode.WORD)
		lb.set_markup(f"<tt>{GLib.markup_escape_text(error_msg)}</tt>")
		# lb.set_label(error_msg)
		# lb.set_selectable(True)
		popup.set_extra_child(lb)

		# Connections
		self.toast.connect("button-clicked", lambda *_: popup.present(self.main_window))
		popup.connect("response", on_response)



from gi.repository import Adw, Gtk, Gio
from src.host_info import HostInfo
import gettext

_ = gettext.gettext


class FilterRow(Adw.ActionRow):
	__gtype_name__ = "FilterRow"

	def __init__(self, item=None, installation=None, **kwargs):
		super().__init__(**kwargs)
		self.item = item
		self.installation = installation
		self.check_button = Gtk.CheckButton()
		self.add_suffix(self.check_button)
		self.set_activatable_widget(self.check_button)


@Gtk.Template(resource_path="/io/github/flattool/Warehouse/packages_page/filters_page.ui")
class FiltersPage(Adw.NavigationPage):
	__gtype_name__ = "FiltersPage"
	gtc = Gtk.Template.Child
	app_check = gtc()
	runtime_check = gtc()
	remotes_group = gtc()
	all_remotes_switch = gtc()
	runtimes_group = gtc()
	all_runtimes_switch = gtc()
	action_bar = gtc()
	reset_button = gtc()

	remote_rows = []
	runtime_rows = []

	def reset_filters(self):
		self.settings.reset("show-apps")
		self.settings.reset("show-runtimes")
		self.settings.reset("remotes-list")
		self.settings.reset("runtimes-list")
		self.generate_filters()
		self.packages_page.apply_filters()

	def is_defaulted(self):
		default = True
		if not self.app_check.get_active():
			default = False
		if self.runtime_check.get_active():
			default = False
		if self.all_remotes_switch.get_active():
			default = False
		if self.all_runtimes_switch.get_active():
			default = False
		self.action_bar.set_revealed(not default)

	def update_gsettings(self):
		self.is_defaulted()
		if not self.is_settings_settable:
			return
		self.settings.set_boolean("show-apps", self.show_apps)
		self.settings.set_boolean("show-runtimes", self.show_runtimes)
		self.settings.set_string("remotes-list", self.remotes_string)
		self.settings.set_string("runtimes-list", self.runtimes_string)
		self.packages_page.apply_filters()

	def app_check_handler(self, *args):
		self.show_apps = self.app_check.get_active()
		self.update_gsettings()

	def runtime_check_handler(self, *args):
		self.show_runtimes = self.runtime_check.get_active()
		self.update_gsettings()

	def all_remotes_handler(self, switch, state):
		self.remotes_string = ""
		if not state:
			self.remotes_string = "all"

		for row in self.remote_rows:
			row.set_visible(state)
			if state and row.check_button.get_active():
				self.remotes_string += f"{row.item.name}<>{row.installation};"
			elif state:
				self.remotes_string.replace(f"{row.item.name}<>{row.installation};", "")

		self.update_gsettings()

	def all_runtimes_handler(self, switch, state):
		self.runtimes_string = ""
		if not state:
			self.runtimes_string = "all"

		for row in self.runtime_rows:
			row.set_visible(state)
			if state and row.check_button.get_active():
				self.runtimes_string += f"{row.item};"
			elif state:
				self.runtimes_string.replace(f"{row.item};", "")

		self.update_gsettings()

	def remote_row_check_handler(self, row):
		if row.check_button.get_active():
			self.remotes_string += f"{row.item.name}<>{row.installation};"
		else:
			self.remotes_string = self.remotes_string.replace(f"{row.item.name}<>{row.installation};", "")
		self.update_gsettings()

	def runtime_row_check_handler(self, row):
		if row.check_button.get_active():
			self.runtimes_string += f"{row.item};"
		else:
			self.runtimes_string = self.runtimes_string.replace(f"{row.item};", "")
		self.update_gsettings()

	def generate_remote_filters(self):
		for row in self.remote_rows:
			self.remotes_group.remove(row)

		self.remote_rows.clear()
		for installation, remotes in HostInfo.remotes.items():
			for remote in remotes:
				if remote.disabled:
					continue

				row = FilterRow(remote, installation)
				row.set_title(remote.title)
				row.set_subtitle(_("Installation: {}").format(installation))
				row.check_button.set_active(f"{remote.name}<>{installation}" in self.remotes_string)
				row.check_button.connect("toggled", lambda *_, row=row: self.remote_row_check_handler(row))
				row.set_visible(self.all_remotes_switch.get_active())
				self.remote_rows.append(row)
				self.remotes_group.add(row)

		self.remotes_group.set_visible(len(self.remote_rows) > 1)
		self.all_remotes_switch.set_active("all" != self.remotes_string)

	def generate_runtime_filters(self):
		for row in self.runtime_rows:
			self.runtimes_group.remove(row)
		self.runtime_rows.clear()
		if len(HostInfo.dependent_runtime_refs) < 2:
			self.runtimes_group.set_visible(False)
			if self.runtimes_string != "all":
				self.runtimes_string = "all"
				self.settings.set_string("runtimes-list", self.runtimes_string)
				self.packages_page.apply_filters()

			return

		for j, ref in enumerate(HostInfo.dependent_runtime_refs):
			row = FilterRow(ref)
			row.set_title(ref)
			row.check_button.set_active(ref in self.runtimes_string)
			row.check_button.connect("toggled", lambda *_, row=row: self.runtime_row_check_handler(row))
			row.set_visible(self.all_runtimes_switch.get_active())
			self.runtime_rows.append(row)
			self.runtimes_group.add(row)

		self.runtimes_group.set_visible(len(self.runtime_rows) > 1)
		self.all_runtimes_switch.set_active("all" != self.runtimes_string)

	def generate_filters(self):
		self.is_settings_settable = False
		self.show_apps = self.settings.get_boolean("show-apps")
		self.show_runtimes = self.settings.get_boolean("show-runtimes")
		self.remotes_string = self.settings.get_string("remotes-list")
		self.runtimes_string = self.settings.get_string("runtimes-list")

		self.app_check.set_active(self.show_apps)
		self.runtime_check.set_active(self.show_runtimes)

		self.generate_remote_filters()
		self.generate_runtime_filters()

		self.is_settings_settable = True

	def __init__(self, **kwargs):
		super().__init__(**kwargs)

		# Extra Objects Creation
		self.packages_page = None  # To be set in packages page
		self.main_window = HostInfo.main_window
		self.settings = Gio.Settings.new("io.github.flattool.Warehouse.filter")
		self.is_settings_settable = False
		self.show_apps = self.settings.get_boolean("show-apps")
		self.show_runtimes = self.settings.get_boolean("show-runtimes")
		self.remotes_string = self.settings.get_string("remotes-list")
		self.runtimes_string = self.settings.get_string("runtimes-list")

		# Apply
		if "," in self.runtimes_string:
			# Convert Warehouse 1.X runtimes filter string from , to ; for item seperationg
			self.runtimes_string = self.runtimes_string.replace(",", ";")
			self.settings.set_string("runtimes-list", self.runtimes_string)

		# Connections
		self.app_check.connect("toggled", self.app_check_handler)
		self.runtime_check.connect("toggled", self.runtime_check_handler)
		self.all_remotes_switch.connect("state-set", self.all_remotes_handler)
		self.all_runtimes_switch.connect("state-set", self.all_runtimes_handler)
		self.reset_button.connect("clicked", lambda *_: self.reset_filters())



from gi.repository import Gio, GLib
from src.host_info import HostInfo
import subprocess, re
import gettext

_ = gettext.gettext


class PackageInstallWorker:
	"""Expect Package Installation Request Data to be Formatted as Such
	[
		{
			"remote": "<remote name>" or "local_file",
			"installation": "<installation name>",
			"package_names": ["<pkg id 1>", "<pkg id 2>", ...],
			"extra_flags": ["<flag 1>", "<flag 2>", ...],
		},
		{
			...
		},
	]
	"""

	groups = None
	process = None
	callback = None
	error_callback = None
	loading_status = None
	total_groups = 0
	cancelled = False

	@classmethod
	def update_status(this, index, package_ratio, complete, total):
		group_ratio = (package_ratio + complete) / (total or 1)
		final_ratio = (group_ratio + index) / (this.total_groups or 1)

		if not this.loading_status is None:
			GLib.idle_add(lambda *_: this.loading_status.progress_bar.set_fraction(final_ratio))

	@classmethod
	def install_thread(this):
		try:
			errors = []
			for index, group in enumerate(this.groups):
				if this.cancelled:
					return

				real_installation = ""
				installation = group["installation"]
				if installation == "user" or installation == "system":
					real_installation = f"--{installation}"
				else:
					real_installation = f"--installation={installation}"

				cmd = ["flatpak", "install", "-y"]

				# Handle local file installs. They don't have a remote specified
				if group["remote"] != "local_file":
					cmd.append(group["remote"])

				cmd += [real_installation] + group["package_names"] + group["extra_flags"]
				this.process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
				percent_pattern = r"\d{1,3}%"
				amount_pattern = r"(\d+)/(\d+)"
				for line in this.process.stdout:
					line = line.strip()
					percent_match = re.search(percent_pattern, line)
					if percent_match:
						ratio = int(percent_match.group()[0:-1]) / 100.0
						amount_match = re.search(amount_pattern, line)
						if amount_match:
							amount = amount_match.group().split("/")
							complete = int(amount[0]) - 1
							total = int(amount[1])
							this.update_status(index, ratio, complete, total)
						else:
							this.update_status(index, ratio, 0, 1)

				this.process.wait(timeout=10)
				if error := this.process.communicate()[1].strip():
					errors.append(error)

			if len(errors) > 0:
				this.on_error(_("Errors occurred during installation"), "\n".join(errors))

		except subprocess.TimeoutExpired:
			this.process.terminate()
			this.on_error(_("Error occurred during installation"), _("Failed to exit cleanly"))

		except Exception as e:
			this.process.terminate()
			this.on_error(_("Error occurred during installation"), str(e))

	@classmethod
	def cancel(this):
		if this.process is None:
			return

		try:
			this.cancelled = True
			this.process.terminate()
			this.process.wait(timeout=10)
		except Exception as e:
			this.on_error(_("Could not cancel installation"), str(e))

	@classmethod
	def on_done(this, *args):
		this.process = None
		HostInfo.main_window.remove_refresh_lockout("installing packages")
		if not this.loading_status is None:
			this.loading_status.progress_bar.set_fraction(0.0)

		if not this.callback is None:
			this.callback()

		this.cancelled = False

	@classmethod
	def on_error(this, user_facing_label, error_message):
		if not this.error_callback is None:
			this.error_callback(user_facing_label, error_message)

	@classmethod
	def install(this, groups, loading_status=None, callback=None, error_callback=None):
		if not this.process is None:
			this.on_error(_("Could not install packages"), _("Packages are currently being installed."))
			return False

		this.callback = callback
		this.groups = groups
		this.total_groups = len(groups)
		this.loading_status = loading_status
		this.error_callback = error_callback

		if this.total_groups < 1:
			this.on_error(_("Could not install packages"), _("No packages were requested to be installed."))
			return False

		HostInfo.main_window.add_refresh_lockout("installing packages")
		Gio.Task.new(None, None, this.on_done).run_in_thread(lambda *_: this.install_thread())
		return True
